# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from starlette.applications import Starlette
from starlette.background import BackgroundTasks
from starlette.exceptions import HTTPException
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from testing_support.asgi_testing import AsgiTest

from newrelic.api.function_trace import FunctionTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_names import callable_name

try:
    from starlette.middleware import Middleware
except ImportError:
    Middleware = None


class HandledError(Exception):
    pass


class NonAsyncHandledError(Exception):
    pass


async def index(request):
    return PlainTextResponse("Hello, world!")


def non_async(request):
    assert current_transaction()
    return PlainTextResponse("Non async route")


async def runtime_error(request):
    raise RuntimeError("Runtime error")


async def handled_error(request):
    raise HandledError("Handled error")


def non_async_handled_error(request):
    raise NonAsyncHandledError("Non async handled error")


async def async_error_handler(request, exc):
    return PlainTextResponse("Async error handler", status_code=500)


def non_async_error_handler(request, exc):
    return PlainTextResponse("Non async error handler", status_code=500)


async def teapot(request):
    raise HTTPException(418, "I'm a teapot")


async def teapot_handler(request, exc):
    return PlainTextResponse("Teapot handler", status_code=418)


def missing_route_handler(request, exc):
    return PlainTextResponse("Missing route handler", status_code=404)


class CustomRoute(object):
    def __init__(self, route):
        self.route = route

    async def __call__(self, scope, receive, send):
        await send({"type": "http.response.start", "status": 200, "headers": []})
        with FunctionTrace(name=callable_name(self.route)):
            await self.route(None)


async def run_bg_task(request):
    tasks = BackgroundTasks()
    tasks.add_task(bg_task_async)
    tasks.add_task(bg_task_non_async)
    return PlainTextResponse("Hello, world!", background=tasks)


async def bg_task_async():
    pass


def bg_task_non_async():
    pass


routes = [
    Route("/index", index),
    Route("/418", teapot),
    Route("/non_async", non_async),
    Route("/runtime_error", runtime_error),
    Route("/handled_error", handled_error),
    Route("/non_async_handled_error", non_async_handled_error),
    Route("/raw_runtime_error", CustomRoute(runtime_error)),
    Route("/raw_http_error", CustomRoute(teapot)),
    Route("/run_bg_task", run_bg_task),
]


def middleware_factory(app):
    async def middleware(scope, receive, send):
        if scope["path"] == "/crash_me_now":
            raise ValueError("Value error")
        return await app(scope, receive, send)

    return middleware


async def middleware_decorator(request, call_next):
    return await call_next(request)


# Generating target applications
app_name_map = {
    "no_error_handler": (True, False, {}),
    "async_error_handler_no_middleware": (
        False,
        False,
        {Exception: async_error_handler},
    ),
    "non_async_error_handler_no_middleware": (False, False, {}),
    "no_middleware": (False, False, {}),
    "debug_no_middleware": (False, True, {}),
    "teapot_exception_handler_no_middleware": (False, False, {}),
}


target_application = dict()
for app_name, flags in app_name_map.items():
    # Bind options
    middleware_on, debug, exception_handlers = flags

    # Instantiate app
    if not middleware_on:
        app = Starlette(
            debug=debug, routes=routes, exception_handlers=exception_handlers
        )
    else:
        if Middleware:
            app = Starlette(
                debug=debug,
                routes=routes,
                middleware=[Middleware(middleware_factory)],
                exception_handlers=exception_handlers,
            )
        else:
            app = Starlette(
                debug=debug, routes=routes, exception_handlers=exception_handlers
            )
            # in earlier versions of starlette, middleware is not a legal argument on the Starlette application class
            # In order to keep the counts the same, we add the middleware twice using the add_middleware interface
            app.add_middleware(middleware_factory)

        app.add_middleware(middleware_factory)
        app.middleware("http")(middleware_decorator)

    # Adding custom exception handlers
    app.add_exception_handler(HandledError, async_error_handler)

    # Add exception handler multiple times to verify the handler is not double wrapped
    app.add_exception_handler(NonAsyncHandledError, non_async_error_handler)
    app.add_exception_handler(NonAsyncHandledError, non_async_error_handler)

    if app_name == "non_async_error_handler_no_middleware":
        app.add_exception_handler(Exception, non_async_error_handler)
        app.add_exception_handler(404, missing_route_handler)
    elif app_name == "teapot_exception_handler_no_middleware":
        app.add_exception_handler(418, teapot_handler)

    # Assign to dict
    target_application[app_name] = AsgiTest(app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture(autouse=True)
def sample_data(collection):
    for x in range(1, 6):
        collection.add({"x": x})

    subcollection_doc = collection.document("subcollection")
    subcollection_doc.set({})
    subcollection_doc.collection("subcollection1").add({})


# ===== AsyncQuery =====


@pytest.fixture()
def exercise_async_query(async_collection):
    async def _exercise_async_query():
        async_query = (
            async_collection.select("x").limit(10).order_by("x").where(field_path="x", op_string="<=", value=3)
        )
        assert len(await async_query.get()) == 3
        assert len([_ async for _ in async_query.stream()]) == 3

    return _exercise_async_query


def test_firestore_async_query(loop, exercise_async_query, async_collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/stream" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/get" % async_collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    # @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_query",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_query")
    def _test():
        loop.run_until_complete(exercise_async_query())

    _test()


@background_task()
def test_firestore_async_query_generators(async_collection, assert_trace_for_async_generator):
    async_query = async_collection.select("x").where(field_path="x", op_string="<=", value=3)
    assert_trace_for_async_generator(async_query.stream)


def test_firestore_async_query_trace_node_datastore_params(loop, exercise_async_query, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_query())

    _test()


# ===== AsyncAggregationQuery =====


@pytest.fixture()
def exercise_async_aggregation_query(async_collection):
    async def _exercise_async_aggregation_query():
        async_aggregation_query = async_collection.select("x").where(field_path="x", op_string="<=", value=3).count()
        assert (await async_aggregation_query.get())[0][0].value == 3
        assert [_ async for _ in async_aggregation_query.stream()][0][0].value == 3

    return _exercise_async_aggregation_query


def test_firestore_async_aggregation_query(loop, exercise_async_aggregation_query, async_collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/stream" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/get" % async_collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_aggregation_query",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_aggregation_query")
    def _test():
        loop.run_until_complete(exercise_async_aggregation_query())

    _test()


@background_task()
def test_firestore_async_aggregation_query_generators(async_collection, assert_trace_for_async_generator):
    async_aggregation_query = async_collection.select("x").where(field_path="x", op_string="<=", value=3).count()
    assert_trace_for_async_generator(async_aggregation_query.stream)


def test_firestore_async_aggregation_query_trace_node_datastore_params(
    loop, exercise_async_aggregation_query, instance_info
):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_aggregation_query())

    _test()


# ===== CollectionGroup =====


@pytest.fixture()
def patch_partition_queries(monkeypatch, async_client, collection, sample_data):
    """
    Partitioning is not implemented in the Firestore emulator.

    Ordinarily this method would return a coroutine that returns an async_generator of Cursor objects.
    Each Cursor must point at a valid document path. To test this, we can patch the RPC to return 1 Cursor
    which is pointed at any document available. The get_partitions will take that and make 2 QueryPartition
    objects out of it, which should be enough to ensure we can exercise the generator's tracing.
    """
    from google.cloud.firestore_v1.types.document import Value
    from google.cloud.firestore_v1.types.query import Cursor

    subcollection = collection.document("subcollection").collection("subcollection1")
    documents = [d for d in subcollection.list_documents()]

    async def mock_partition_query(*args, **kwargs):
        async def _mock_partition_query():
            yield Cursor(before=False, values=[Value(reference_value=documents[0].path)])

        return _mock_partition_query()

    monkeypatch.setattr(async_client._firestore_api, "partition_query", mock_partition_query)
    yield


@pytest.fixture()
def exercise_async_collection_group(async_client, async_collection):
    async def _exercise_async_collection_group():
        async_collection_group = async_client.collection_group(async_collection.id)
        assert len(await async_collection_group.get())
        assert len([d async for d in async_collection_group.stream()])

        partitions = [p async for p in async_collection_group.get_partitions(1)]
        assert len(partitions) == 2
        documents = []
        while partitions:
            documents.extend(await partitions.pop().query().get())
        assert len(documents) == 6

    return _exercise_async_collection_group


def test_firestore_async_collection_group(
    loop, exercise_async_collection_group, async_collection, patch_partition_queries, instance_info
):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/get" % async_collection.id, 3),
        ("Datastore/statement/Firestore/%s/stream" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/get_partitions" % async_collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 3),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/operation/Firestore/get_partitions", 1),
        ("Datastore/all", 5),
        ("Datastore/allOther", 5),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 5),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_collection_group",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_collection_group")
    def _test():
        loop.run_until_complete(exercise_async_collection_group())

    _test()


@background_task()
def test_firestore_async_collection_group_generators(
    async_client, async_collection, assert_trace_for_async_generator, patch_partition_queries
):
    async_collection_group = async_client.collection_group(async_collection.id)
    assert_trace_for_async_generator(async_collection_group.get_partitions, 1)


def test_firestore_async_collection_group_trace_node_datastore_params(
    loop, exercise_async_collection_group, instance_info, patch_partition_queries
):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_collection_group())

    _test()


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from testing_support.fixtures import validate_attributes

from newrelic.api.background_task import background_task
from newrelic.hooks.mlmodel_sklearn import PredictReturnTypeProxy


@pytest.mark.parametrize(
    "metric_scorer_name",
    (
        "accuracy_score",
        "balanced_accuracy_score",
        "f1_score",
        "precision_score",
        "recall_score",
        "roc_auc_score",
        "r2_score",
    ),
)
def test_metric_scorer_attributes(metric_scorer_name, run_metric_scorer):
    @validate_attributes("agent", ["DecisionTreeClassifier/TrainingStep/0/%s" % metric_scorer_name])
    @background_task()
    def _test():
        run_metric_scorer(metric_scorer_name)

    _test()


@pytest.mark.parametrize(
    "metric_scorer_name",
    (
        "accuracy_score",
        "balanced_accuracy_score",
        "f1_score",
        "precision_score",
        "recall_score",
        "roc_auc_score",
        "r2_score",
    ),
)
def test_metric_scorer_training_steps_attributes(metric_scorer_name, run_metric_scorer):
    @validate_attributes(
        "agent",
        [
            "DecisionTreeClassifier/TrainingStep/0/%s" % metric_scorer_name,
            "DecisionTreeClassifier/TrainingStep/1/%s" % metric_scorer_name,
        ],
    )
    @background_task()
    def _test():
        run_metric_scorer(metric_scorer_name, training_steps=[0, 1])

    _test()


@pytest.mark.parametrize(
    "metric_scorer_name,kwargs",
    [
        ("f1_score", {"average": None}),
        ("precision_score", {"average": None}),
        ("recall_score", {"average": None}),
    ],
)
def test_metric_scorer_iterable_score_attributes(metric_scorer_name, kwargs, run_metric_scorer):
    @validate_attributes(
        "agent",
        [
            "DecisionTreeClassifier/TrainingStep/0/%s[0]" % metric_scorer_name,
            "DecisionTreeClassifier/TrainingStep/0/%s[1]" % metric_scorer_name,
        ],
    )
    @background_task()
    def _test():
        run_metric_scorer(metric_scorer_name, kwargs)

    _test()


@pytest.mark.parametrize(
    "metric_scorer_name",
    [
        "accuracy_score",
        "balanced_accuracy_score",
        "f1_score",
        "precision_score",
        "recall_score",
        "roc_auc_score",
        "r2_score",
    ],
)
def test_metric_scorer_attributes_unknown_model(metric_scorer_name):
    @validate_attributes("agent", ["Unknown/TrainingStep/Unknown/%s" % metric_scorer_name])
    @background_task()
    def _test():
        from sklearn import metrics

        y_pred = [1, 0]
        y_test = [1, 0]

        getattr(metrics, metric_scorer_name)(y_test, y_pred)

    _test()


@pytest.mark.parametrize("data", (np.array([0, 1]), "foo", 1, 1.0, True, [0, 1], {"foo": "bar"}, (0, 1), np.str_("F")))
def test_PredictReturnTypeProxy(data):
    wrapped_data = PredictReturnTypeProxy(data, "ModelName", 0)

    assert wrapped_data._nr_model_name == "ModelName"
    assert wrapped_data._nr_training_step == 0


@pytest.fixture
def run_metric_scorer():
    def _run(metric_scorer_name, metric_scorer_kwargs=None, training_steps=None):
        from sklearn import metrics, tree

        x_train = [[0, 0], [1, 1]]
        y_train = [0, 1]
        x_test = [[2.0, 2.0], [0, 0.5]]
        y_test = [1, 0]

        if not training_steps:
            training_steps = [0]

        clf = tree.DecisionTreeClassifier(random_state=0)
        for step in training_steps:
            model = clf.fit(x_train, y_train)

            labels = model.predict(x_test)

            metric_scorer_kwargs = metric_scorer_kwargs or {}
            getattr(metrics, metric_scorer_name)(y_test, labels, **metric_scorer_kwargs)

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611

from newrelic.common.object_wrapper import (  # noqa: F401 pylint: disable=W0611
    transient_function_wrapper,
)


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (framework_sanic)", default_settings=_default_settings
)


RESPONSES = []

loop = None


def create_request_class(app, method, url, headers=None, loop=None):
    from sanic.request import Request

    try:
        _request = Request(
            method=method.upper(),
            url_bytes=url.encode("utf-8"),
            headers=headers,
            version="1.0",
            transport=None,
        )
    except TypeError:
        _request = Request(
            app=app,
            method=method.upper(),
            url_bytes=url.encode("utf-8"),
            headers=headers,
            version="1.0",
            transport=None,
        )

    try:
        # Manually initialize HTTP protocol
        from sanic.http import Http, Stage
        from sanic.server import HttpProtocol

        class MockProtocol(HttpProtocol):
            async def send(*args, **kwargs):  # pylint: disable=E0211
                return

        proto = MockProtocol(loop=loop, app=app)
        proto.recv_buffer = bytearray()
        http = Http(proto)

        if hasattr(http, "init_for_request"):
            http.init_for_request()

        http.stage = Stage.HANDLER
        http.response_func = http.http1_response_header
        _request.stream = http
    except ImportError:
        pass

    return _request


def create_request_coroutine(app, method, url, headers=None, loop=None):
    headers = headers or {}
    try:
        coro = app.handle_request(create_request_class(app, method, url, headers, loop=loop))
    except TypeError:

        def write_callback(response):
            response.raw_headers = response.output()
            if b"write_response_error" in response.raw_headers:
                raise ValueError("write_response_error")

            if response not in RESPONSES:
                RESPONSES.append(response)

        async def stream_callback(response):
            response.raw_headers = response.get_headers()
            if response not in RESPONSES:
                RESPONSES.append(response)

        coro = app.handle_request(
            create_request_class(app, method, url, headers, loop=loop),
            write_callback,
            stream_callback,
        )

    return coro


def request(app, method, url, headers=None):
    global loop
    if loop is None:
        loop = asyncio.new_event_loop()

    if not getattr(app.router, "finalized", True):
        # Handle startup if the router hasn't been finalized.
        # Older versions don't have this requirement or variable so
        # the default should be True.
        if hasattr(app, "_startup"):
            loop.run_until_complete(app._startup())
        else:
            app.router.finalize()
    # Starting in 22.9.0 sanic introduced an API to control middleware ordering.
    # This included a new method called finalize_middleware that sets the middleware
    # to be used on the request.route during the app._startup. In order to register
    # new middleware the finalize_middleware must be called.
    elif hasattr(app, "finalize_middleware"):
        app.finalize_middleware()

    coro = create_request_coroutine(app, method, url, headers, loop)
    loop.run_until_complete(coro)
    return RESPONSES.pop()


class TestApplication(object):
    def __init__(self, app):
        self.app = app

    def fetch(self, method, url, headers=None):
        return request(self.app, method, url, headers)


@pytest.fixture()
def app():
    from _target_application import app

    return TestApplication(app)


@pytest.fixture(autouse=True)
def capture_requests(monkeypatch):
    from sanic.response import BaseHTTPResponse

    original = BaseHTTPResponse.__init__

    def capture(*args, **kwargs):
        original(*args, **kwargs)
        response = args[0]
        if getattr(response, "status", None) is None:
            response.status = 200

        if response not in RESPONSES:
            RESPONSES.append(response)

    monkeypatch.setattr(BaseHTTPResponse, "__init__", capture)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import (
    override_application_settings,
    reset_core_stats_engine,
    validate_error_event_sample_data,
)
from testing_support.validators.validate_error_event_attributes_outside_transaction import (
    validate_error_event_attributes_outside_transaction,
)
from testing_support.validators.validate_error_trace_attributes_outside_transaction import (
    validate_error_trace_attributes_outside_transaction,
)
from testing_support.validators.validate_time_metrics_outside_transaction import (
    validate_time_metrics_outside_transaction,
)
from testing_support.validators.validate_transaction_error_trace_attributes import (
    validate_transaction_error_trace_attributes,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import current_transaction
from newrelic.common.object_names import callable_name

_runtime_error_name = callable_name(RuntimeError)
_error_message = "Test error message."

# Settings presets

# Error classes settings
expected_runtime_error_settings = {"error_collector.expected_classes": [_runtime_error_name]}
ignore_runtime_error_settings = {"error_collector.ignore_classes": [_runtime_error_name]}

# Status code settings
expected_status_code_settings = {"error_collector.expected_status_codes": [418]}
ignore_status_code_settings = {"error_collector.ignore_status_codes": [418]}

_test_runtime_error = [(_runtime_error_name, _error_message)]
_intrinsic_attributes = {
    "error.class": _runtime_error_name,
    "error.message": _error_message,
    "error.expected": False,
    "transactionName": "OtherTransaction/Function/test",
}

# =============== Test ignored/expected classes ===============

classes_settings_matrix = [
    ({}, False, False),
    (ignore_runtime_error_settings, False, True),
    (expected_runtime_error_settings, True, False),
]
override_expected_matrix = (True, False, None)


def exercise(expected=None, ignore=None, status_code=None, application=None):
    try:
        raise RuntimeError(_error_message)
    except RuntimeError:
        if current_transaction() is None:
            # Record exception inside transaction
            application = application or application_instance()

        notice_error(
            expected=expected,
            ignore=ignore,
            status_code=status_code,
            application=application,
        )


@pytest.mark.parametrize("settings,expected,ignore", classes_settings_matrix)
def test_classes_error_event_inside_transaction(settings, expected, ignore):
    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected

    error_count = 1 if not ignore else 0
    errors = _test_runtime_error if not ignore else []
    expected_errors = _runtime_error_name if expected and not ignore else None

    @validate_transaction_errors(errors=errors, expected_errors=expected_errors)
    @validate_error_event_sample_data(
        required_attrs=attributes,
        required_user_attrs=False,
        num_errors=error_count,
    )
    @background_task(name="test")
    @override_application_settings(settings)
    def _test():
        exercise()

    _test()


@pytest.mark.parametrize("settings,expected,ignore", classes_settings_matrix)
def test_classes_error_event_outside_transaction(settings, expected, ignore):
    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected
    attributes["transactionName"] = None

    error_count = 1 if not ignore else 0

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        exact_attrs={"intrinsic": attributes, "agent": {}, "user": {}},
        num_errors=error_count,
    )
    @override_application_settings(settings)
    def _test():
        exercise()

    _test()


# =============== Test error trace attributes ===============
# Override testing is required to ensure the proper handling of
# overrides. First attempt at implementation did not correctly handle this
# and resulted in discrepancies between error events and traced errors.


error_trace_settings_matrix = [
    ({}, False),
    (expected_runtime_error_settings, True),
]
override_expected_matrix = (True, False, None)


@pytest.mark.parametrize("settings,expected", error_trace_settings_matrix)
@pytest.mark.parametrize("override_expected", override_expected_matrix)
def test_error_trace_attributes_inside_transaction(settings, expected, override_expected):
    expected = override_expected if override_expected is not None else expected

    error_trace_attributes = {
        "intrinsic": {
            "error.expected": expected,
        },
        "agent": {},
        "user": {},
    }

    @validate_transaction_error_trace_attributes(exact_attrs=error_trace_attributes)
    @background_task(name="test")
    @override_application_settings(settings)
    def _test():
        exercise(override_expected)

    _test()


@pytest.mark.parametrize("settings,expected", error_trace_settings_matrix)
@pytest.mark.parametrize("override_expected", override_expected_matrix)
def test_error_trace_attributes_outside_transaction(settings, expected, override_expected):
    expected = override_expected if override_expected is not None else expected

    error_trace_attributes = {
        "intrinsic": {
            "error.class": _runtime_error_name,
            "error.message": _error_message,
            "error.expected": expected,
            "transactionName": None,
        },
        "agent": {},
        "user": {},
    }

    @reset_core_stats_engine()
    @validate_error_trace_attributes_outside_transaction(_runtime_error_name, exact_attrs=error_trace_attributes)
    @override_application_settings(settings)
    def _test():
        exercise(override_expected)

    _test()


# =============== Test metrics not incremented ===============


@pytest.mark.parametrize("expected", override_expected_matrix)
def test_error_metrics_inside_transaction(expected):
    normal_metrics_count = None if expected else 1
    expected_metrics_count = 1 if expected else None
    metrics_payload = [
        ("Errors/all", normal_metrics_count),
        ("Errors/OtherTransaction/Function/test", normal_metrics_count),
        ("Errors/allOther", normal_metrics_count),
        ("ErrorsExpected/all", expected_metrics_count),
    ]

    @validate_transaction_metrics("test", background_task=True, rollup_metrics=metrics_payload)
    @background_task(name="test")
    def _test():
        exercise(expected)

    _test()


@pytest.mark.parametrize("expected", override_expected_matrix)
def test_error_metrics_outside_transaction(expected):
    normal_metrics_count = None if expected else 1
    expected_metrics_count = 1 if expected else None
    metrics_payload = [
        ("Errors/all", normal_metrics_count),
        ("ErrorsExpected/all", expected_metrics_count),
    ]

    @reset_core_stats_engine()
    @validate_time_metrics_outside_transaction(metrics_payload)
    def _test():
        exercise(expected)

    _test()


# =============== Test ignored/expected status codes ===============


class TeapotError(RuntimeError):
    status_code = 418


_teapot_error_name = callable_name(TeapotError)
_test_teapot_error = [(_teapot_error_name, _error_message)]


def retrieve_status_code(exc, value, tb):
    return value.status_code


status_codes_settings_matrix = [
    ({}, False, False),
    (ignore_status_code_settings, False, True),
    (expected_status_code_settings, True, False),
]

status_code_matrix = [None, 418, retrieve_status_code]


@pytest.mark.parametrize("settings,expected,ignore", status_codes_settings_matrix)
@pytest.mark.parametrize("status_code", status_code_matrix)
def test_status_codes_inside_transaction(settings, expected, ignore, status_code):
    if status_code is None:
        # Override all settings
        ignore = False
        expected = False

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected
    attributes["error.class"] = _teapot_error_name

    error_count = 1 if not ignore else 0
    errors = _test_teapot_error if not ignore else []
    expected_errors = _teapot_error_name if expected and not ignore else None

    @validate_transaction_errors(errors=errors, expected_errors=expected_errors)
    @validate_error_event_sample_data(
        required_attrs=attributes,
        required_user_attrs=False,
        num_errors=error_count,
    )
    @background_task(name="test")
    @override_application_settings(settings)
    def _test():
        try:
            raise TeapotError(_error_message)
        except:
            notice_error(status_code=status_code)

    _test()


@pytest.mark.parametrize("settings,expected,ignore", status_codes_settings_matrix)
@pytest.mark.parametrize("status_code", status_code_matrix)
def test_status_codes_outside_transaction(settings, expected, ignore, status_code):
    if status_code is None:
        # Override all settings
        ignore = False
        expected = False

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected
    attributes["error.class"] = _teapot_error_name
    attributes["transactionName"] = None

    error_count = 1 if not ignore else 0

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        exact_attrs={"intrinsic": attributes, "agent": {}, "user": {}},
        num_errors=error_count,
    )
    @override_application_settings(settings)
    def _test():
        try:
            raise TeapotError(_error_message)
        except:
            notice_error(status_code=status_code, application=application_instance(activate=False))

    _test()


# =============== Test mixed ignored and expected settings ===============

# Aggregate settings
ignore_status_code_expected_class_settings = {}
ignore_status_code_expected_class_settings.update(ignore_status_code_settings)
ignore_status_code_expected_class_settings.update(expected_runtime_error_settings)
expected_status_code_ignore_class_settings = {}
expected_status_code_ignore_class_settings.update(expected_status_code_settings)
expected_status_code_ignore_class_settings.update(ignore_runtime_error_settings)
combined_runtime_error_settings = {}
combined_runtime_error_settings.update(expected_runtime_error_settings)
combined_runtime_error_settings.update(ignore_runtime_error_settings)
combined_status_code_settings = {}
combined_status_code_settings.update(expected_status_code_settings)
combined_status_code_settings.update(ignore_status_code_settings)

mixed_settings_matrix = [
    ({}, False, False),
    (ignore_status_code_expected_class_settings, True, True),
    (expected_status_code_ignore_class_settings, True, True),
    (combined_runtime_error_settings, True, True),
    (combined_status_code_settings, True, True),
]
override_expected_matrix = (True, False, None)
override_ignore_matrix = (True, False, None)


@pytest.mark.parametrize("settings,expected,ignore", mixed_settings_matrix)
@pytest.mark.parametrize("override_expected", override_expected_matrix)
@pytest.mark.parametrize("override_ignore", override_ignore_matrix)
def test_mixed_ignore_expected_settings_inside_transaction(
    settings, expected, ignore, override_expected, override_ignore
):
    expected = override_expected if override_expected is not None else expected
    ignore = override_ignore if override_ignore is not None else ignore

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected

    error_count = 1 if not ignore else 0
    errors = _test_runtime_error if not ignore else []
    expected_errors = _runtime_error_name if expected and not ignore else None

    @validate_transaction_errors(errors=errors, expected_errors=expected_errors)
    @validate_error_event_sample_data(
        required_attrs=attributes,
        required_user_attrs=False,
        num_errors=error_count,
    )
    @background_task(name="test")
    @override_application_settings(settings)
    def _test():
        exercise(override_expected, override_ignore, status_code=418)

    _test()


@pytest.mark.parametrize("settings,expected,ignore", mixed_settings_matrix)
@pytest.mark.parametrize("override_expected", override_expected_matrix)
@pytest.mark.parametrize("override_ignore", override_ignore_matrix)
def test_mixed_ignore_expected_settings_outside_transaction(
    settings, expected, ignore, override_expected, override_ignore
):
    expected = override_expected if override_expected is not None else expected
    ignore = override_ignore if override_ignore is not None else ignore

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected
    attributes["transactionName"] = None

    error_count = 1 if not ignore else 0

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        exact_attrs={"intrinsic": attributes, "agent": {}, "user": {}},
        num_errors=error_count,
    )
    @override_application_settings(settings)
    def _test():
        exercise(override_expected, override_ignore, status_code=418)

    _test()


# =============== Test multiple override types ===============

override_matrix = (
    (True, True),
    (False, False),
    (lambda e, v, t: True, True),
    (lambda e, v, t: False, False),
    ([_runtime_error_name], True),
    ([], False),
    (None, False),
)


@pytest.mark.parametrize("override,result", override_matrix)
@pytest.mark.parametrize("parameter", ("ignore", "expected"))
def test_overrides_inside_transaction(override, result, parameter):
    ignore = result if parameter == "ignore" else False
    expected = result if parameter == "expected" else False
    kwargs = {parameter: override}

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected

    error_count = 1 if not ignore else 0
    errors = _test_runtime_error if not ignore else []
    expected_errors = _runtime_error_name if expected and not ignore else None

    @validate_transaction_errors(errors=errors, expected_errors=expected_errors)
    @validate_error_event_sample_data(
        required_attrs=attributes,
        required_user_attrs=False,
        num_errors=error_count,
    )
    @background_task(name="test")
    def _test():
        exercise(**kwargs)

    _test()


@pytest.mark.parametrize("override,result", override_matrix)
@pytest.mark.parametrize("parameter", ("ignore", "expected"))
def test_overrides_outside_transaction(override, result, parameter):
    ignore = result if parameter == "ignore" else False
    expected = result if parameter == "expected" else False
    kwargs = {parameter: override}

    # Update attributes with parameters
    attributes = _intrinsic_attributes.copy()
    attributes["error.expected"] = expected
    attributes["transactionName"] = None

    error_count = 1 if not ignore else 0

    @reset_core_stats_engine()
    @validate_error_event_attributes_outside_transaction(
        exact_attrs={"intrinsic": attributes, "agent": {}, "user": {}},
        num_errors=error_count,
    )
    def _test():
        exercise(**kwargs)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import sys
import time

import pytest
from testing_support.fixtures import capture_transaction_metrics
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import database_trace
from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import external_trace
from newrelic.api.function_trace import function_trace
from newrelic.api.graphql_trace import graphql_operation_trace, graphql_resolver_trace
from newrelic.api.memcache_trace import memcache_trace
from newrelic.api.message_trace import message_trace


@pytest.mark.parametrize(
    "trace,metric",
    [
        (functools.partial(function_trace, name="simple_gen"), "Function/simple_gen"),
        (functools.partial(external_trace, library="lib", url="http://foo.com"), "External/foo.com/lib/"),
        (functools.partial(database_trace, "select * from foo"), "Datastore/statement/None/foo/select"),
        (functools.partial(datastore_trace, "lib", "foo", "bar"), "Datastore/statement/lib/foo/bar"),
        (functools.partial(message_trace, "lib", "op", "typ", "name"), "MessageBroker/lib/typ/op/Named/name"),
        (functools.partial(memcache_trace, "cmd"), "Memcache/cmd"),
        (functools.partial(graphql_operation_trace), "GraphQL/operation/GraphQL/<unknown>/<anonymous>/<unknown>"),
        (functools.partial(graphql_resolver_trace), "GraphQL/resolve/GraphQL/<unknown>"),
    ],
)
def test_awaitable_timing(event_loop, trace, metric):
    @trace()
    async def coro():
        await asyncio.sleep(0.1)

    @background_task(name="test_awaitable")
    async def parent():
        await coro()

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_awaitable", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    def _test():
        event_loop.run_until_complete(parent())

    _test()

    # Check that coroutines time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.1


@pytest.mark.skipif(sys.version_info >= (3, 11), reason="Asyncio decorator was removed in Python 3.11+.")
@pytest.mark.parametrize(
    "trace,metric",
    [
        (functools.partial(function_trace, name="simple_gen"), "Function/simple_gen"),
        (functools.partial(external_trace, library="lib", url="http://foo.com"), "External/foo.com/lib/"),
        (functools.partial(database_trace, "select * from foo"), "Datastore/statement/None/foo/select"),
        (functools.partial(datastore_trace, "lib", "foo", "bar"), "Datastore/statement/lib/foo/bar"),
        (functools.partial(message_trace, "lib", "op", "typ", "name"), "MessageBroker/lib/typ/op/Named/name"),
        (functools.partial(memcache_trace, "cmd"), "Memcache/cmd"),
        (functools.partial(graphql_operation_trace), "GraphQL/operation/GraphQL/<unknown>/<anonymous>/<unknown>"),
        (functools.partial(graphql_resolver_trace), "GraphQL/resolve/GraphQL/<unknown>"),
    ],
)
@pytest.mark.parametrize("yield_from", [True, False])
@pytest.mark.parametrize("use_await", [True, False])
@pytest.mark.parametrize("coro_decorator_first", [True, False])
def test_asyncio_decorator_timing(event_loop, trace, metric, yield_from, use_await, coro_decorator_first):

    if yield_from:

        def coro():
            yield from asyncio.sleep(0.1)

    else:

        def coro():
            time.sleep(0.1)

    if coro_decorator_first:
        coro = trace()(asyncio.coroutine(coro))
    else:
        coro = asyncio.coroutine(trace()(coro))

    if use_await:

        async def parent():
            await coro()

    else:

        @asyncio.coroutine
        def parent():
            yield from coro()

    parent = background_task(name="test_awaitable")(parent)

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_awaitable", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    def _test():
        event_loop.run_until_complete(parent())

    _test()

    # Check that coroutines time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.1



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from newrelic.core.config import global_settings
from newrelic.core.stack_trace import (exception_stack, current_stack,
    _format_stack_trace as _format_stack_trace_from_dicts, _extract_stack,
    _extract_tb)


def _format_stack_trace_from_tuples(frames):
    result = ['Traceback (most recent call last):']
    result.extend(['File "{0}", line {1}, in {2}'.format(*v) for v in frames])
    return result

def function0():
    raise RuntimeError('function0')

def function1():
    function0()

def function2():
    function1()

def function3():
    function2()

def function4():
    function3()

def function5():
    function4()

def function6():
    function5()

def function7():
    function6()

def function8():
    function7()

def function9():
    function8()

def function10():
    function9()

def function11():
    function10()

def function12():
    function11()

# When using a try/except with stack trace formatting being done within
# the except block, all line numbers for the combine stack trace will be
# correct.

def test_trace_inline():
    try:
        function12()
    except Exception:
        tb = sys.exc_info()[2]
        actual = exception_stack(tb, limit=14)
        require = _stack_trace_inline
        assert actual == require, (actual, require)

_stack_trace_inline = _format_stack_trace_from_tuples([
    (__file__, test_trace_inline.__code__.co_firstlineno+2,
        test_trace_inline.__name__),
    (__file__, function12.__code__.co_firstlineno+1, function12.__name__),
    (__file__, function11.__code__.co_firstlineno+1, function11.__name__),
    (__file__, function10.__code__.co_firstlineno+1, function10.__name__),
    (__file__, function9.__code__.co_firstlineno+1, function9.__name__),
    (__file__, function8.__code__.co_firstlineno+1, function8.__name__),
    (__file__, function7.__code__.co_firstlineno+1, function7.__name__),
    (__file__, function6.__code__.co_firstlineno+1, function6.__name__),
    (__file__, function5.__code__.co_firstlineno+1, function5.__name__),
    (__file__, function4.__code__.co_firstlineno+1, function4.__name__),
    (__file__, function3.__code__.co_firstlineno+1, function3.__name__),
    (__file__, function2.__code__.co_firstlineno+1, function2.__name__),
    (__file__, function1.__code__.co_firstlineno+1, function1.__name__),
    (__file__, function0.__code__.co_firstlineno+1, function0.__name__)
])

# Where the traceback is saved away and then returned to a scope outside
# of the except block, the line numbers for the exception part of the
# stack are correct, but things go funny with what is joined from the
# current stack. In particular, the line numbers of parent stack frames
# can be wrong as what the traceback holds is a reference to the live
# stack frames and since the execution point of the stack frames changes
# so does the line number.
#
# The consequence of this is that rather than the line number for the
# stack frame for test_trace_passed1() being line 1, it is line 2,
# which is where the parent stack frame is being calculated within the
# exception_stack() function.

def _test_trace_passed1():
    try:
        function12()
    except Exception:
        return sys.exc_info()[2]

def test_trace_passed1():
    tb = _test_trace_passed1()
    actual = exception_stack(tb, limit=15)
    require = _stack_trace_passed1
    assert actual == require, (actual, require)

_stack_trace_passed1 = _format_stack_trace_from_tuples([
    (__file__, test_trace_passed1.__code__.co_firstlineno+2,
        test_trace_passed1.__name__),
    (__file__, _test_trace_passed1.__code__.co_firstlineno+2,
        _test_trace_passed1.__name__),
    (__file__, function12.__code__.co_firstlineno+1, function12.__name__),
    (__file__, function11.__code__.co_firstlineno+1, function11.__name__),
    (__file__, function10.__code__.co_firstlineno+1, function10.__name__),
    (__file__, function9.__code__.co_firstlineno+1, function9.__name__),
    (__file__, function8.__code__.co_firstlineno+1, function8.__name__),
    (__file__, function7.__code__.co_firstlineno+1, function7.__name__),
    (__file__, function6.__code__.co_firstlineno+1, function6.__name__),
    (__file__, function5.__code__.co_firstlineno+1, function5.__name__),
    (__file__, function4.__code__.co_firstlineno+1, function4.__name__),
    (__file__, function3.__code__.co_firstlineno+1, function3.__name__),
    (__file__, function2.__code__.co_firstlineno+1, function2.__name__),
    (__file__, function1.__code__.co_firstlineno+1, function1.__name__),
    (__file__, function0.__code__.co_firstlineno+1, function0.__name__)
])

# An additional example for the changing line numbers as described above
# is the following. Note that _test_trace_passed2b() does not actually
# appear. This is correct, as the intersection point for the stack trace
# is test_trace_passed2() and the line number shows the call to the
# function _test_trace_passed2b() where the code is excuting within that
# frame. This is instead of being at the point of call for the function
# _test_trace_passed2a() where the exception was generated.

def _test_trace_passed2a():
    try:
        function12()
    except Exception:
        return sys.exc_info()[2]

def _test_trace_passed2b(tb):
    actual = exception_stack(tb, limit=15)
    require = _stack_trace_passed2
    assert actual == require, (actual, require)

def test_trace_passed2():
    tb = _test_trace_passed2a()
    _test_trace_passed2b(tb)

_stack_trace_passed2 = _format_stack_trace_from_tuples([
    (__file__, test_trace_passed2.__code__.co_firstlineno+2,
        test_trace_passed2.__name__),
    (__file__, _test_trace_passed2a.__code__.co_firstlineno+2,
        _test_trace_passed2a.__name__),
    (__file__, function12.__code__.co_firstlineno+1, function12.__name__),
    (__file__, function11.__code__.co_firstlineno+1, function11.__name__),
    (__file__, function10.__code__.co_firstlineno+1, function10.__name__),
    (__file__, function9.__code__.co_firstlineno+1, function9.__name__),
    (__file__, function8.__code__.co_firstlineno+1, function8.__name__),
    (__file__, function7.__code__.co_firstlineno+1, function7.__name__),
    (__file__, function6.__code__.co_firstlineno+1, function6.__name__),
    (__file__, function5.__code__.co_firstlineno+1, function5.__name__),
    (__file__, function4.__code__.co_firstlineno+1, function4.__name__),
    (__file__, function3.__code__.co_firstlineno+1, function3.__name__),
    (__file__, function2.__code__.co_firstlineno+1, function2.__name__),
    (__file__, function1.__code__.co_firstlineno+1, function1.__name__),
    (__file__, function0.__code__.co_firstlineno+1, function0.__name__)
])

# Here we limit to bottom most stack frames within just the exception
# stack.

def test_trace_truncated():
    try:
        function12()
    except Exception:
        tb = sys.exc_info()[2]
        actual = exception_stack(tb, limit=5)
        require = _stack_trace_limit_truncated
        assert actual == require, (actual, require)

_stack_trace_limit_truncated = _format_stack_trace_from_tuples([
    (__file__, function4.__code__.co_firstlineno+1, function4.__name__),
    (__file__, function3.__code__.co_firstlineno+1, function3.__name__),
    (__file__, function2.__code__.co_firstlineno+1, function2.__name__),
    (__file__, function1.__code__.co_firstlineno+1, function1.__name__),
    (__file__, function0.__code__.co_firstlineno+1, function0.__name__)
])

# Previous examples truncated at stack frame of test as can't easily
# compare to frames above. This test collects all stack frames to make
# sure it can handle them, but can't compare.

def test_trace_exception_full():
    try:
        function12()
    except Exception:
        tb = sys.exc_info()[2]
        actual = exception_stack(tb)


def test_trace_line_limit():
    settings = global_settings()
    original = settings.max_stack_trace_lines
    settings.max_stack_trace_lines = 40
    try:
        function12()
    except Exception:
        tb = sys.exc_info()[2]
        actual = exception_stack(tb)
    finally:
        settings.max_stack_trace_lines = original

    # 40 lines of stack trace and one line that says
    # 'Traceback (most recent call last):'

    assert len(actual) == 41

# Test just the ability to get the current stack frame. Limited so we
# can compare to what we expect.

def skip0(skip, limit):
    return current_stack(skip=skip, limit=limit)

def skip1(skip, limit):
    return skip0(skip, limit)

def skip2(skip, limit):
    return skip1(skip, limit)

def skip3(skip, limit):
    return skip2(skip, limit)

def skip4(skip, limit):
    return skip3(skip, limit)

def skip5(skip, limit):
    return skip4(skip, limit)

def test_trace_current():
    actual = skip5(skip=0, limit=5)
    require = _stack_trace_current
    assert actual == require, (actual, require)

_stack_trace_current = _format_stack_trace_from_tuples([
    (__file__, skip4.__code__.co_firstlineno+1, skip4.__name__),
    (__file__, skip3.__code__.co_firstlineno+1, skip3.__name__),
    (__file__, skip2.__code__.co_firstlineno+1, skip2.__name__),
    (__file__, skip1.__code__.co_firstlineno+1, skip1.__name__),
    (__file__, skip0.__code__.co_firstlineno+1, skip0.__name__)
])

# Test the ability to skip some frames at the bottom of the stack.

def test_trace_current_skip():
    actual = skip5(skip=1, limit=5)
    require = _stack_trace_current_skip
    assert actual == require, (actual, require)

_stack_trace_current_skip = _format_stack_trace_from_tuples([
    (__file__, skip5.__code__.co_firstlineno+1, skip5.__name__),
    (__file__, skip4.__code__.co_firstlineno+1, skip4.__name__),
    (__file__, skip3.__code__.co_firstlineno+1, skip3.__name__),
    (__file__, skip2.__code__.co_firstlineno+1, skip2.__name__),
    (__file__, skip1.__code__.co_firstlineno+1, skip1.__name__)
])

# Previous examples truncated current stack as can't easily compare to
# frames above. This test collects all stack frames to make sure it can
# handle them, but can't compare.

def test_trace_current_full():
    actual = skip5(skip=0, limit=1000)

# This example shows the relationship between functions _extract_stack()
# and _extract_tb(). _extract_stack() collects the outer frames starting
# from the frame above where the exception occurred. _extract_tb()
# collects the inner frames from the try/except block down to the
# exception. Call them explicitly and the result of adding the two
# should be the same as having used the exception_stack() function. This
# test just takes the limit calculations done inside of the function
# exception_stack() out of play.

def test_trace_current_plus_traceback():
    try:
        raise ZeroDivisionError
    except ZeroDivisionError:
        tb = sys.exc_info()[2]
        frame = tb.tb_frame.f_back

    limit = 10000

    actual = exception_stack(tb)
    require = _format_stack_trace_from_dicts(
            _extract_stack(frame, 0, limit)+_extract_tb(tb, limit))

    assert actual == require, (actual, require)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.application import application_instance as __application
from newrelic.api.application import application_settings as __application_settings
from newrelic.api.application import register_application as __register_application
from newrelic.api.log import NewRelicContextFormatter  # noqa
from newrelic.api.time_trace import (
    add_custom_span_attribute as __add_custom_span_attribute,
)
from newrelic.api.time_trace import current_trace as __current_trace
from newrelic.api.time_trace import get_linking_metadata as __get_linking_metadata
from newrelic.api.time_trace import notice_error as __notice_error
from newrelic.api.time_trace import record_exception as __record_exception
from newrelic.api.transaction import (
    accept_distributed_trace_headers as __accept_distributed_trace_headers,
)
from newrelic.api.transaction import (
    accept_distributed_trace_payload as __accept_distributed_trace_payload,
)
from newrelic.api.transaction import add_custom_attribute as __add_custom_attribute
from newrelic.api.transaction import add_custom_attributes as __add_custom_attributes
from newrelic.api.transaction import add_custom_parameter as __add_custom_parameter
from newrelic.api.transaction import add_custom_parameters as __add_custom_parameters
from newrelic.api.transaction import add_framework_info as __add_framework_info
from newrelic.api.transaction import capture_request_params as __capture_request_params
from newrelic.api.transaction import (
    create_distributed_trace_payload as __create_distributed_trace_payload,
)
from newrelic.api.transaction import current_span_id as __current_span_id
from newrelic.api.transaction import current_trace_id as __current_trace_id
from newrelic.api.transaction import current_transaction as __current_transaction
from newrelic.api.transaction import (
    disable_browser_autorum as __disable_browser_autorum,
)
from newrelic.api.transaction import end_of_transaction as __end_of_transaction
from newrelic.api.transaction import (
    get_browser_timing_footer as __get_browser_timing_footer,
)
from newrelic.api.transaction import (
    get_browser_timing_header as __get_browser_timing_header,
)
from newrelic.api.transaction import ignore_transaction as __ignore_transaction
from newrelic.api.transaction import (
    insert_distributed_trace_headers as __insert_distributed_trace_headers,
)
from newrelic.api.transaction import record_custom_event as __record_custom_event
from newrelic.api.transaction import record_custom_metric as __record_custom_metric
from newrelic.api.transaction import record_custom_metrics as __record_custom_metrics
from newrelic.api.transaction import record_log_event as __record_log_event
from newrelic.api.transaction import record_ml_event as __record_ml_event
from newrelic.api.transaction import set_background_task as __set_background_task
from newrelic.api.transaction import set_transaction_name as __set_transaction_name
from newrelic.api.transaction import suppress_apdex_metric as __suppress_apdex_metric
from newrelic.api.transaction import (
    suppress_transaction_trace as __suppress_transaction_trace,
)
from newrelic.api.wsgi_application import (
    WSGIApplicationWrapper as __WSGIApplicationWrapper,
)
from newrelic.api.wsgi_application import (
    wrap_wsgi_application as __wrap_wsgi_application,
)
from newrelic.api.wsgi_application import wsgi_application as __wsgi_application
from newrelic.config import extra_settings as __extra_settings
from newrelic.config import initialize as __initialize
from newrelic.core.agent import register_data_source as __register_data_source
from newrelic.core.agent import shutdown_agent as __shutdown_agent
from newrelic.core.config import global_settings as __global_settings
from newrelic.samplers.decorators import data_source_factory as __data_source_factory
from newrelic.samplers.decorators import (
    data_source_generator as __data_source_generator,
)

try:
    from newrelic.api.asgi_application import (
        ASGIApplicationWrapper as __ASGIApplicationWrapper,
    )
    from newrelic.api.asgi_application import asgi_application as __asgi_application
    from newrelic.api.asgi_application import (
        wrap_asgi_application as __wrap_asgi_application,
    )
except SyntaxError:

    def __asgi_application(*args, **kwargs):
        pass

    __ASGIApplicationWrapper = __asgi_application
    __wrap_asgi_application = __asgi_application

from newrelic.api.background_task import BackgroundTask as __BackgroundTask
from newrelic.api.background_task import (
    BackgroundTaskWrapper as __BackgroundTaskWrapper,
)
from newrelic.api.background_task import background_task as __background_task
from newrelic.api.background_task import wrap_background_task as __wrap_background_task
from newrelic.api.database_trace import DatabaseTrace as __DatabaseTrace
from newrelic.api.database_trace import DatabaseTraceWrapper as __DatabaseTraceWrapper
from newrelic.api.database_trace import database_trace as __database_trace
from newrelic.api.database_trace import (
    register_database_client as __register_database_client,
)
from newrelic.api.database_trace import wrap_database_trace as __wrap_database_trace
from newrelic.api.datastore_trace import DatastoreTrace as __DatastoreTrace
from newrelic.api.datastore_trace import (
    DatastoreTraceWrapper as __DatastoreTraceWrapper,
)
from newrelic.api.datastore_trace import datastore_trace as __datastore_trace
from newrelic.api.datastore_trace import wrap_datastore_trace as __wrap_datastore_trace
from newrelic.api.error_trace import ErrorTrace as __ErrorTrace
from newrelic.api.error_trace import ErrorTraceWrapper as __ErrorTraceWrapper
from newrelic.api.error_trace import error_trace as __error_trace
from newrelic.api.error_trace import wrap_error_trace as __wrap_error_trace
from newrelic.api.external_trace import ExternalTrace as __ExternalTrace
from newrelic.api.external_trace import ExternalTraceWrapper as __ExternalTraceWrapper
from newrelic.api.external_trace import external_trace as __external_trace
from newrelic.api.external_trace import wrap_external_trace as __wrap_external_trace
from newrelic.api.function_trace import FunctionTrace as __FunctionTrace
from newrelic.api.function_trace import FunctionTraceWrapper as __FunctionTraceWrapper
from newrelic.api.function_trace import function_trace as __function_trace
from newrelic.api.function_trace import wrap_function_trace as __wrap_function_trace
from newrelic.api.generator_trace import (
    GeneratorTraceWrapper as __GeneratorTraceWrapper,
)
from newrelic.api.generator_trace import generator_trace as __generator_trace
from newrelic.api.generator_trace import wrap_generator_trace as __wrap_generator_trace
from newrelic.api.html_insertion import insert_html_snippet as __insert_html_snippet
from newrelic.api.html_insertion import verify_body_exists as __verify_body_exists
from newrelic.api.lambda_handler import LambdaHandlerWrapper as __LambdaHandlerWrapper
from newrelic.api.lambda_handler import lambda_handler as __lambda_handler
from newrelic.api.message_trace import MessageTrace as __MessageTrace
from newrelic.api.message_trace import MessageTraceWrapper as __MessageTraceWrapper
from newrelic.api.message_trace import message_trace as __message_trace
from newrelic.api.message_trace import wrap_message_trace as __wrap_message_trace
from newrelic.api.message_transaction import MessageTransaction as __MessageTransaction
from newrelic.api.message_transaction import (
    MessageTransactionWrapper as __MessageTransactionWrapper,
)
from newrelic.api.message_transaction import (
    message_transaction as __message_transaction,
)
from newrelic.api.message_transaction import (
    wrap_message_transaction as __wrap_message_transaction,
)
from newrelic.api.ml_model import wrap_mlmodel as __wrap_mlmodel
from newrelic.api.profile_trace import ProfileTraceWrapper as __ProfileTraceWrapper
from newrelic.api.profile_trace import profile_trace as __profile_trace
from newrelic.api.profile_trace import wrap_profile_trace as __wrap_profile_trace
from newrelic.api.settings import set_error_group_callback as __set_error_group_callback
from newrelic.api.supportability import wrap_api_call as __wrap_api_call
from newrelic.api.transaction import set_user_id as __set_user_id
from newrelic.api.transaction_name import (
    TransactionNameWrapper as __TransactionNameWrapper,
)
from newrelic.api.transaction_name import transaction_name as __transaction_name
from newrelic.api.transaction_name import (
    wrap_transaction_name as __wrap_transaction_name,
)
from newrelic.api.web_transaction import WebTransaction as __WebTransaction
from newrelic.api.web_transaction import (
    WebTransactionWrapper as __WebTransactionWrapper,
)
from newrelic.api.web_transaction import web_transaction as __web_transaction
from newrelic.api.web_transaction import wrap_web_transaction as __wrap_web_transaction
from newrelic.common.object_names import callable_name as __callable_name
from newrelic.common.object_wrapper import FunctionWrapper as __FunctionWrapper
from newrelic.common.object_wrapper import InFunctionWrapper as __InFunctionWrapper
from newrelic.common.object_wrapper import ObjectProxy as __ObjectProxy
from newrelic.common.object_wrapper import ObjectWrapper as __ObjectWrapper
from newrelic.common.object_wrapper import OutFunctionWrapper as __OutFunctionWrapper
from newrelic.common.object_wrapper import PostFunctionWrapper as __PostFunctionWrapper
from newrelic.common.object_wrapper import PreFunctionWrapper as __PreFunctionWrapper
from newrelic.common.object_wrapper import function_wrapper as __function_wrapper
from newrelic.common.object_wrapper import in_function as __in_function
from newrelic.common.object_wrapper import out_function as __out_function
from newrelic.common.object_wrapper import (
    patch_function_wrapper as __patch_function_wrapper,
)
from newrelic.common.object_wrapper import post_function as __post_function
from newrelic.common.object_wrapper import pre_function as __pre_function
from newrelic.common.object_wrapper import resolve_path as __resolve_path
from newrelic.common.object_wrapper import (
    transient_function_wrapper as __transient_function_wrapper,
)
from newrelic.common.object_wrapper import (
    wrap_function_wrapper as __wrap_function_wrapper,
)
from newrelic.common.object_wrapper import wrap_in_function as __wrap_in_function
from newrelic.common.object_wrapper import wrap_object as __wrap_object
from newrelic.common.object_wrapper import (
    wrap_object_attribute as __wrap_object_attribute,
)
from newrelic.common.object_wrapper import wrap_out_function as __wrap_out_function
from newrelic.common.object_wrapper import wrap_post_function as __wrap_post_function
from newrelic.common.object_wrapper import wrap_pre_function as __wrap_pre_function

# EXPERIMENTAL - Generator traces are currently experimental and may not
# exist in this form in future versions of the agent.

initialize = __initialize
extra_settings = __wrap_api_call(__extra_settings, "extra_settings")
global_settings = __wrap_api_call(__global_settings, "global_settings")
shutdown_agent = __wrap_api_call(__shutdown_agent, "shutdown_agent")
register_data_source = __wrap_api_call(__register_data_source, "register_data_source")
data_source_generator = __wrap_api_call(__data_source_generator, "data_source_generator")
data_source_factory = __wrap_api_call(__data_source_factory, "data_source_factory")
application = __wrap_api_call(__application, "application")
register_application = __register_application
application_settings = __wrap_api_call(__application_settings, "application_settings")
current_trace = __wrap_api_call(__current_trace, "current_trace")
get_linking_metadata = __wrap_api_call(__get_linking_metadata, "get_linking_metadata")
add_custom_span_attribute = __wrap_api_call(__add_custom_span_attribute, "add_custom_span_attribute")
current_transaction = __wrap_api_call(__current_transaction, "current_transaction")
set_user_id = __wrap_api_call(__set_user_id, "set_user_id")
set_error_group_callback = __wrap_api_call(__set_error_group_callback, "set_error_group_callback")
set_transaction_name = __wrap_api_call(__set_transaction_name, "set_transaction_name")
end_of_transaction = __wrap_api_call(__end_of_transaction, "end_of_transaction")
set_background_task = __wrap_api_call(__set_background_task, "set_background_task")
ignore_transaction = __wrap_api_call(__ignore_transaction, "ignore_transaction")
suppress_apdex_metric = __wrap_api_call(__suppress_apdex_metric, "suppress_apdex_metric")
capture_request_params = __wrap_api_call(__capture_request_params, "capture_request_params")
add_custom_parameter = __wrap_api_call(__add_custom_parameter, "add_custom_parameter")
add_custom_parameters = __wrap_api_call(__add_custom_parameters, "add_custom_parameters")
add_custom_attribute = __wrap_api_call(__add_custom_attribute, "add_custom_attribute")
add_custom_attributes = __wrap_api_call(__add_custom_attributes, "add_custom_attributes")
add_framework_info = __wrap_api_call(__add_framework_info, "add_framework_info")
record_exception = __wrap_api_call(__record_exception, "record_exception")
notice_error = __wrap_api_call(__notice_error, "notice_error")
get_browser_timing_header = __wrap_api_call(__get_browser_timing_header, "get_browser_timing_header")
get_browser_timing_footer = __wrap_api_call(__get_browser_timing_footer, "get_browser_timing_footer")
disable_browser_autorum = __wrap_api_call(__disable_browser_autorum, "disable_browser_autorum")
suppress_transaction_trace = __wrap_api_call(__suppress_transaction_trace, "suppress_transaction_trace")
record_custom_metric = __wrap_api_call(__record_custom_metric, "record_custom_metric")
record_custom_metrics = __wrap_api_call(__record_custom_metrics, "record_custom_metrics")
record_custom_event = __wrap_api_call(__record_custom_event, "record_custom_event")
record_log_event = __wrap_api_call(__record_log_event, "record_log_event")
record_ml_event = __wrap_api_call(__record_ml_event, "record_ml_event")
accept_distributed_trace_payload = __wrap_api_call(
    __accept_distributed_trace_payload, "accept_distributed_trace_payload"
)
create_distributed_trace_payload = __wrap_api_call(
    __create_distributed_trace_payload, "create_distributed_trace_payload"
)
accept_distributed_trace_headers = __wrap_api_call(
    __accept_distributed_trace_headers, "accept_distributed_trace_headers"
)
insert_distributed_trace_headers = __wrap_api_call(
    __insert_distributed_trace_headers, "insert_distributed_trace_headers"
)
current_trace_id = __wrap_api_call(__current_trace_id, "current_trace_id")
current_span_id = __wrap_api_call(__current_span_id, "current_span_id")
wsgi_application = __wsgi_application
asgi_application = __asgi_application
WebTransaction = __wrap_api_call(__WebTransaction, "WebTransaction")
web_transaction = __wrap_api_call(__web_transaction, "web_transaction")
WebTransactionWrapper = __wrap_api_call(__WebTransactionWrapper, "WebTransactionWrapper")
wrap_web_transaction = __wrap_api_call(__wrap_web_transaction, "wrap_web_transaction")
WSGIApplicationWrapper = __WSGIApplicationWrapper
wrap_wsgi_application = __wrap_wsgi_application
ASGIApplicationWrapper = __ASGIApplicationWrapper
wrap_asgi_application = __wrap_asgi_application
background_task = __wrap_api_call(__background_task, "background_task")
BackgroundTask = __wrap_api_call(__BackgroundTask, "BackgroundTask")
BackgroundTaskWrapper = __wrap_api_call(__BackgroundTaskWrapper, "BackgroundTaskWrapper")
wrap_background_task = __wrap_api_call(__wrap_background_task, "wrap_background_task")
LambdaHandlerWrapper = __wrap_api_call(__LambdaHandlerWrapper, "LambdaHandlerWrapper")
lambda_handler = __wrap_api_call(__lambda_handler, "lambda_handler")
transaction_name = __wrap_api_call(__transaction_name, "transaction_name")
TransactionNameWrapper = __wrap_api_call(__TransactionNameWrapper, "TransactionNameWrapper")
wrap_transaction_name = __wrap_api_call(__wrap_transaction_name, "wrap_transaction_name")
function_trace = __wrap_api_call(__function_trace, "function_trace")
FunctionTrace = __wrap_api_call(__FunctionTrace, "FunctionTrace")
FunctionTraceWrapper = __wrap_api_call(__FunctionTraceWrapper, "FunctionTraceWrapper")
wrap_function_trace = __wrap_api_call(__wrap_function_trace, "wrap_function_trace")
generator_trace = __wrap_api_call(__generator_trace, "generator_trace")
GeneratorTraceWrapper = __wrap_api_call(__GeneratorTraceWrapper, "GeneratorTraceWrapper")
wrap_generator_trace = __wrap_api_call(__wrap_generator_trace, "wrap_generator_trace")
profile_trace = __wrap_api_call(__profile_trace, "profile_trace")
ProfileTraceWrapper = __wrap_api_call(__ProfileTraceWrapper, "ProfileTraceWrapper")
wrap_profile_trace = __wrap_api_call(__wrap_profile_trace, "wrap_profile_trace")
database_trace = __wrap_api_call(__database_trace, "database_trace")
DatabaseTrace = __wrap_api_call(__DatabaseTrace, "DatabaseTrace")
DatabaseTraceWrapper = __wrap_api_call(__DatabaseTraceWrapper, "DatabaseTraceWrapper")
wrap_database_trace = __wrap_api_call(__wrap_database_trace, "wrap_database_trace")
register_database_client = __wrap_api_call(__register_database_client, "register_database_client")
datastore_trace = __wrap_api_call(__datastore_trace, "datastore_trace")
DatastoreTrace = __wrap_api_call(__DatastoreTrace, "DatastoreTrace")
DatastoreTraceWrapper = __wrap_api_call(__DatastoreTraceWrapper, "DatastoreTraceWrapper")
wrap_datastore_trace = __wrap_api_call(__wrap_datastore_trace, "wrap_datastore_trace")
external_trace = __wrap_api_call(__external_trace, "external_trace")
ExternalTrace = __wrap_api_call(__ExternalTrace, "ExternalTrace")
ExternalTraceWrapper = __wrap_api_call(__ExternalTraceWrapper, "ExternalTraceWrapper")
wrap_external_trace = __wrap_api_call(__wrap_external_trace, "wrap_external_trace")
error_trace = __wrap_api_call(__error_trace, "error_trace")
ErrorTrace = __wrap_api_call(__ErrorTrace, "ErrorTrace")
ErrorTraceWrapper = __wrap_api_call(__ErrorTraceWrapper, "ErrorTraceWrapper")
wrap_error_trace = __wrap_api_call(__wrap_error_trace, "wrap_error_trace")
message_trace = __wrap_api_call(__message_trace, "message_trace")
MessageTrace = __wrap_api_call(__MessageTrace, "MessageTrace")
MessageTraceWrapper = __wrap_api_call(__MessageTraceWrapper, "MessageTraceWrapper")
wrap_message_trace = __wrap_api_call(__wrap_message_trace, "wrap_message_trace")
message_transaction = __wrap_api_call(__message_transaction, "message_trace")
MessageTransaction = __wrap_api_call(__MessageTransaction, "MessageTransaction")
MessageTransactionWrapper = __wrap_api_call(__MessageTransactionWrapper, "MessageTransactionWrapper")
wrap_message_transaction = __wrap_api_call(__wrap_message_transaction, "wrap_message_transaction")
callable_name = __wrap_api_call(__callable_name, "callable_name")
ObjectProxy = __wrap_api_call(__ObjectProxy, "ObjectProxy")
wrap_object = __wrap_api_call(__wrap_object, "wrap_object")
wrap_object_attribute = __wrap_api_call(__wrap_object_attribute, "wrap_object_attribute")
resolve_path = __wrap_api_call(__resolve_path, "resolve_path")
transient_function_wrapper = __wrap_api_call(__transient_function_wrapper, "transient_function_wrapper")
FunctionWrapper = __wrap_api_call(__FunctionWrapper, "FunctionWrapper")
function_wrapper = __wrap_api_call(__function_wrapper, "function_wrapper")
wrap_function_wrapper = __wrap_api_call(__wrap_function_wrapper, "wrap_function_wrapper")
patch_function_wrapper = __wrap_api_call(__patch_function_wrapper, "patch_function_wrapper")
ObjectWrapper = __wrap_api_call(__ObjectWrapper, "ObjectWrapper")
pre_function = __wrap_api_call(__pre_function, "pre_function")
PreFunctionWrapper = __wrap_api_call(__PreFunctionWrapper, "PreFunctionWrapper")
wrap_pre_function = __wrap_api_call(__wrap_pre_function, "wrap_pre_function")
post_function = __wrap_api_call(__post_function, "post_function")
PostFunctionWrapper = __wrap_api_call(__PostFunctionWrapper, "PostFunctionWrapper")
wrap_post_function = __wrap_api_call(__wrap_post_function, "wrap_post_function")
in_function = __wrap_api_call(__in_function, "in_function")
InFunctionWrapper = __wrap_api_call(__InFunctionWrapper, "InFunctionWrapper")
wrap_in_function = __wrap_api_call(__wrap_in_function, "wrap_in_function")
out_function = __wrap_api_call(__out_function, "out_function")
OutFunctionWrapper = __wrap_api_call(__OutFunctionWrapper, "OutFunctionWrapper")
wrap_out_function = __wrap_api_call(__wrap_out_function, "wrap_out_function")
insert_html_snippet = __wrap_api_call(__insert_html_snippet, "insert_html_snippet")
verify_body_exists = __wrap_api_call(__verify_body_exists, "verify_body_exists")
wrap_mlmodel = __wrap_api_call(__wrap_mlmodel, "wrap_mlmodel")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
from testing_support.validators.validate_transaction_count import (
    validate_transaction_count,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.package_version_utils import get_package_version_tuple

starlette_version = get_package_version_tuple("starlette")[:3]

try:
    from starlette.middleware import Middleware  # noqa: F401

    no_middleware = False
except ImportError:
    no_middleware = True

skip_if_no_middleware = pytest.mark.skipif(no_middleware, reason="These tests verify middleware functionality")


@pytest.fixture(scope="session")
def target_application():
    import _test_bg_tasks

    return _test_bg_tasks.target_application


@pytest.mark.parametrize("route", ["async", "sync"])
def test_simple(target_application, route):
    route_metrics = [("Function/_test_bg_tasks:run_%s_bg_task" % route, 1)]

    @validate_transaction_metrics("_test_bg_tasks:run_%s_bg_task" % route, index=-2, scoped_metrics=route_metrics)
    @validate_transaction_metrics("_test_bg_tasks:%s_bg_task" % route, background_task=True)
    @validate_transaction_count(2)
    def _test():
        app = target_application["none"]
        response = app.get("/" + route)
        assert response.status == 200

    _test()


@skip_if_no_middleware
@pytest.mark.parametrize("route", ["async", "sync"])
def test_asgi_style_middleware(target_application, route):
    route_metrics = [("Function/_test_bg_tasks:run_%s_bg_task" % route, 1)]

    @validate_transaction_metrics("_test_bg_tasks:run_%s_bg_task" % route, index=-2, scoped_metrics=route_metrics)
    @validate_transaction_metrics("_test_bg_tasks:%s_bg_task" % route, background_task=True)
    @validate_transaction_count(2)
    def _test():
        app = target_application["asgi"]
        response = app.get("/" + route)
        assert response.status == 200

    _test()


@skip_if_no_middleware
@pytest.mark.parametrize("route", ["async", "sync"])
def test_basehttp_style_middleware(target_application, route):
    route_metric = ("Function/_test_bg_tasks:run_%s_bg_task" % route, 1)
    # A function trace metric that appears only when the bug below is present, causing background tasks to be
    # completed inside web transactions, requiring a function trace to be used for timing
    # instead of a background task transaction. Should not be present at all when bug is fixed.
    bg_task_metric = ("Function/_test_bg_tasks:%s_bg_task" % route, 1)

    def _test():
        app = target_application["basehttp"]
        response = app.get("/" + route)
        assert response.status == 200

    # The bug was fixed in version 0.21.0 but re-occured in 0.23.1.
    # The bug was also not present on 0.20.1 to 0.23.1 if using Python3.7.
    # The bug was fixed again in version 0.29.0
    BUG_COMPLETELY_FIXED = any(
        (
            (0, 21, 0) <= starlette_version < (0, 23, 1),
            (0, 20, 1) <= starlette_version < (0, 23, 1) and sys.version_info[:2] > (3, 7),
            starlette_version >= (0, 29, 0),
        )
    )
    BUG_PARTIALLY_FIXED = any(
        (
            (0, 20, 1) <= starlette_version < (0, 21, 0),
            (0, 23, 1) <= starlette_version < (0, 29, 0),
        )
    )
    if BUG_COMPLETELY_FIXED:
        # Assert both web transaction and background task transactions are present.
        _test = validate_transaction_metrics(
            "_test_bg_tasks:run_%s_bg_task" % route, index=-2, scoped_metrics=[route_metric]
        )(_test)
        _test = validate_transaction_metrics("_test_bg_tasks:%s_bg_task" % route, background_task=True)(_test)
        _test = validate_transaction_count(2)(_test)
    elif BUG_PARTIALLY_FIXED:
        # The background task no longer blocks the completion of the web request/web transaction.
        # However, the BaseHTTPMiddleware causes the task to be cancelled when the web request disconnects, so there are no
        # longer function traces or background task transactions.
        # In version 0.23.1, the check to see if more_body exists is removed, reverting behavior to this model
        _test = validate_transaction_metrics("_test_bg_tasks:run_%s_bg_task" % route, scoped_metrics=[route_metric])(
            _test
        )
        _test = validate_transaction_count(1)(_test)
    else:
        # The BaseHTTPMiddleware causes the background task to execute within the web request
        # with the web transaction still active.
        _test = validate_transaction_metrics(
            "_test_bg_tasks:run_%s_bg_task" % route, scoped_metrics=[route_metric, bg_task_metric]
        )(_test)
        _test = validate_transaction_count(1)(_test)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import sys
import webtest

from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

import cherrypy

class EndPoint(object):

    def index(self):
        return 'INDEX RESPONSE'

dispatcher = cherrypy.dispatch.RoutesDispatcher()
dispatcher.connect(action='index', name='endpoint', route='/endpoint',
            controller=EndPoint())

conf = { '/': { 'request.dispatch': dispatcher } }

application = cherrypy.Application(None, '/', conf)
test_application = webtest.TestApp(application)

@validate_code_level_metrics("test_routes.EndPoint", "index")
@validate_transaction_errors(errors=[])
def test_routes_index():
    response = test_application.get('/endpoint')
    response.mustcontain('INDEX RESPONSE')

@validate_transaction_errors(errors=[])
def test_on_index_agent_disabled():
    environ = { 'newrelic.enabled': False }
    response = test_application.get('/endpoint', extra_environ=environ)
    response.mustcontain('INDEX RESPONSE')

@validate_transaction_errors(errors=[])
def test_routes_not_found():
    response = test_application.get('/missing', status=404)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

INI_FILE_WITHOUT_LICENSE_KEY = b"""
[newrelic]
"""

NO_REGION_KEY = '66c637a29c3982469a3fe8d1982d002c4a'
INI_FILE_NO_REGION_KEY = """
[newrelic]
license_key = %s
""" % NO_REGION_KEY
INI_FILE_NO_REGION_KEY = INI_FILE_NO_REGION_KEY.encode('utf-8')

EU01_KEY = 'eu01xx66c637a29c3982469a3fe8d1982d002c4a'
INI_FILE_EU01_KEY = """
[newrelic]
license_key = %s
""" % EU01_KEY
INI_FILE_EU01_KEY = INI_FILE_EU01_KEY.encode('utf-8')

INI_FILE_HOST_OVERRIDE = """
[newrelic]
host = staging-collector.newrelic.com
license_key = %s
""" % EU01_KEY
INI_FILE_HOST_OVERRIDE = INI_FILE_HOST_OVERRIDE.encode('utf-8')

STAGING_HOST = 'staging-collector.newrelic.com'
STAGING_HOST_ENV = {'NEW_RELIC_HOST': STAGING_HOST}
EVIL_COLLECTOR_HOST_ENV = {'NEW_RELIC_HOST': 'evil-collector.newrelic.com'}

NO_REGION_ENV = {'NEW_RELIC_LICENSE_KEY': NO_REGION_KEY}
EU_REGION_ENV = {'NEW_RELIC_LICENSE_KEY': EU01_KEY}
EU01_HOST = 'collector.eu01.nr-data.net'

DEFAULT_HOST = 'collector.newrelic.com'


# Tests for loading settings and testing for values precedence
@pytest.mark.parametrize('ini,env,expected_host,expected_license_key', [
    # Test license key parsing (no env vars needed since we're not testing
    # precedence)
    (INI_FILE_NO_REGION_KEY, {}, 'collector.newrelic.com', NO_REGION_KEY),
    (INI_FILE_EU01_KEY, {}, 'collector.eu01.nr-data.net', EU01_KEY),

    # Test precedence
    # 1. host in config file (this trumps all)
    (INI_FILE_HOST_OVERRIDE, EVIL_COLLECTOR_HOST_ENV, STAGING_HOST, EU01_KEY),

    # 2. host in environment variable (overrides parsed host)
    (INI_FILE_EU01_KEY, STAGING_HOST_ENV, STAGING_HOST, EU01_KEY),

    # 3. host parsed by license key in config file
    (INI_FILE_EU01_KEY, NO_REGION_ENV, EU01_HOST, EU01_KEY),
    (INI_FILE_NO_REGION_KEY, EU_REGION_ENV, DEFAULT_HOST, NO_REGION_KEY),

    # 4. host parsed by license key in env variable
    (INI_FILE_WITHOUT_LICENSE_KEY, EU_REGION_ENV, EU01_HOST, EU01_KEY),
    (INI_FILE_WITHOUT_LICENSE_KEY, NO_REGION_ENV, DEFAULT_HOST, NO_REGION_KEY),

    # 5. defaults to collector.newrelic.com (no license key specified)
    (INI_FILE_WITHOUT_LICENSE_KEY, {}, DEFAULT_HOST, None),
])
def test_region_aware_license_keys(ini, env, expected_host,
        expected_license_key, global_settings):

    settings = global_settings()
    assert settings.host == expected_host
    assert settings.license_key == expected_license_key


@pytest.mark.parametrize("ini,license_key,host_override,expected_host", (
    (INI_FILE_WITHOUT_LICENSE_KEY, NO_REGION_KEY, None, "collector.newrelic.com"),
    (INI_FILE_WITHOUT_LICENSE_KEY, EU01_KEY, None, "collector.eu01.nr-data.net"),
    (INI_FILE_WITHOUT_LICENSE_KEY, NO_REGION_KEY, "foo.newrelic.com", "foo.newrelic.com"),
))
def test_region_aware_global_settings(ini, license_key, host_override,
        expected_host, global_settings):
    settings = global_settings()

    if host_override:
        settings.host = host_override

    settings.license_key = license_key
    assert settings.host == expected_host



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import elasticsearch
import elasticsearch.client
import pytest
from conftest import ES_VERSION
from testing_support.validators.validate_datastore_trace_inputs import (
    validate_datastore_trace_inputs,
)

from newrelic.api.background_task import background_task

RUN_IF_V8 = pytest.mark.skipif(
    ES_VERSION < (8,), reason="Only run for v8+. We don't support all methods in previous versions."
)


@pytest.fixture
def client(client):
    if ES_VERSION < (8, 0):
        client.index(
            index="contacts", doc_type="person", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1
        )
    else:
        client.index(index="contacts", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    return client


@pytest.mark.parametrize(
    "sub_module,method,args,kwargs,expected_index",
    [
        (None, "exists", (), {"index": "contacts", "id": 1}, "contacts"),
        (None, "info", (), {}, None),
        pytest.param(
            None,
            "msearch",
            (),
            {"searches": [{}, {"query": {"match": {"message": "this is a test"}}}], "index": "contacts"},
            "contacts",
            marks=RUN_IF_V8,
        ),
        ("indices", "exists", (), {"index": "contacts"}, "contacts"),
        ("indices", "exists_template", (), {"name": "no-exist"}, None),
        ("cat", "count", (), {"index": "contacts"}, "contacts"),
        ("cat", "health", (), {}, None),
        pytest.param(
            "cluster",
            "allocation_explain",
            (),
            {"index": "contacts", "shard": 0, "primary": True},
            "contacts",
            marks=RUN_IF_V8,
        ),
        ("cluster", "get_settings", (), {}, None),
        ("cluster", "health", (), {"index": "contacts"}, "contacts"),
        ("nodes", "info", (), {}, None),
        ("snapshot", "status", (), {}, None),
        ("tasks", "list", (), {}, None),
        ("ingest", "geo_ip_stats", (), {}, None),
    ],
)
def test_method_on_client_datastore_trace_inputs(client, sub_module, method, args, kwargs, expected_index):
    expected_operation = "%s.%s" % (sub_module, method) if sub_module else method

    @validate_datastore_trace_inputs(target=expected_index, operation=expected_operation)
    @background_task()
    def _test():
        if not sub_module:
            getattr(client, method)(*args, **kwargs)
        else:
            getattr(getattr(client, sub_module), method)(*args, **kwargs)

    _test()


def _test_methods_wrapped(_object, ignored_methods=None):
    if not ignored_methods:
        ignored_methods = {"perform_request", "transport"}

    def is_wrapped(m):
        return hasattr(getattr(_object, m), "__wrapped__")

    methods = {m for m in dir(_object) if not m[0] == "_"}
    uninstrumented = {m for m in (methods - ignored_methods) if not is_wrapped(m)}
    assert not uninstrumented, "There are uninstrumented methods: %s" % uninstrumented


@RUN_IF_V8
def test_instrumented_methods_client():
    _test_methods_wrapped(elasticsearch.Elasticsearch)


@RUN_IF_V8
def test_instrumented_methods_client_indices():
    _test_methods_wrapped(elasticsearch.client.IndicesClient)


@RUN_IF_V8
def test_instrumented_methods_client_cluster():
    _test_methods_wrapped(elasticsearch.client.ClusterClient)


@RUN_IF_V8
def test_instrumented_methods_client_cat():
    if hasattr(elasticsearch.client, "CatClient"):
        _test_methods_wrapped(elasticsearch.client.CatClient)


@RUN_IF_V8
def test_instrumented_methods_client_nodes():
    if hasattr(elasticsearch.client, "NodesClient"):
        _test_methods_wrapped(elasticsearch.client.NodesClient)


@RUN_IF_V8
def test_instrumented_methods_client_snapshot():
    if hasattr(elasticsearch.client, "SnapshotClient"):
        _test_methods_wrapped(elasticsearch.client.SnapshotClient)


@RUN_IF_V8
def test_instrumented_methods_client_tasks():
    if hasattr(elasticsearch.client, "TasksClient"):
        _test_methods_wrapped(elasticsearch.client.TasksClient)


@RUN_IF_V8
def test_instrumented_methods_client_ingest():
    if hasattr(elasticsearch.client, "IngestClient"):
        _test_methods_wrapped(elasticsearch.client.IngestClient)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.packages import six


def test_model_methods_wrapped_in_function_trace(tree_model_name, run_tree_model):
    # Note: in the following expected metrics, predict and predict_proba are called by
    # score and predict_log_proba so they are expected to be called twice instead of
    # once like the rest of the methods.
    expected_scoped_metrics = {
        "ExtraTreeRegressor": [
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.score", 1),
        ],
        "DecisionTreeClassifier": [
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict_proba", 2),
        ],
        "ExtraTreeClassifier": [
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict_proba", 2),
        ],
        "DecisionTreeRegressor": [
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.score", 1),
        ],
    }
    expected_transaction_name = (
        "test_tree_models:test_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_tree_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[tree_model_name],
        rollup_metrics=expected_scoped_metrics[tree_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_tree_model()

    _test()


def test_multiple_calls_to_model_methods(tree_model_name, run_tree_model):
    # Note: in the following expected metrics, predict and predict_proba are called by
    # score and predict_log_proba so they are expected to be called twice as often as
    # the other methods.
    expected_scoped_metrics = {
        "ExtraTreeRegressor": [
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.predict", 4),
            ("Function/MLModel/Sklearn/Named/ExtraTreeRegressor.score", 2),
        ],
        "DecisionTreeClassifier": [
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict", 4),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.score", 2),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict_log_proba", 2),
            ("Function/MLModel/Sklearn/Named/DecisionTreeClassifier.predict_proba", 4),
        ],
        "ExtraTreeClassifier": [
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict", 4),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.score", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict_log_proba", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreeClassifier.predict_proba", 4),
        ],
        "DecisionTreeRegressor": [
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.predict", 4),
            ("Function/MLModel/Sklearn/Named/DecisionTreeRegressor.score", 2),
        ],
    }
    expected_transaction_name = (
        "test_tree_models:test_multiple_calls_to_model_methods.<locals>._test" if six.PY3 else "test_tree_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[tree_model_name],
        rollup_metrics=expected_scoped_metrics[tree_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        x_test = [[2.0, 2.0], [2.0, 1.0]]
        y_test = [1, 1]

        model = run_tree_model()

        model.predict(x_test)
        model.score(x_test, y_test)
        # Some models don't have these methods.
        if hasattr(model, "predict_log_proba"):
            model.predict_log_proba(x_test)
        if hasattr(model, "predict_proba"):
            model.predict_proba(x_test)

    _test()


@pytest.fixture(params=["ExtraTreeRegressor", "DecisionTreeClassifier", "ExtraTreeClassifier", "DecisionTreeRegressor"])
def tree_model_name(request):
    return request.param


@pytest.fixture
def run_tree_model(tree_model_name):
    def _run():
        import sklearn.tree

        x_train = [[0, 0], [1, 1]]
        y_train = [0, 1]
        x_test = [[2.0, 2.0], [2.0, 1.0]]
        y_test = [1, 1]

        clf = getattr(sklearn.tree, tree_model_name)(random_state=0)
        model = clf.fit(x_train, y_train)

        labels = model.predict(x_test)
        model.score(x_test, y_test)
        # Some models don't have these methods.
        if hasattr(model, "predict_log_proba"):
            model.predict_log_proba(x_test)
        if hasattr(model, "predict_proba"):
            model.predict_proba(x_test)
        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import base64

from testing_support.fixtures import (
    override_ignore_status_codes,
    override_application_settings)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from newrelic.packages import six
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

import webtest

from bottle import __version__ as version

version = [int(x) for x in version.split('-')[0].split('.')]

if len(version) == 2:
    version.append(0)

version = tuple(version)

requires_auth_basic = pytest.mark.skipif(version < (0, 9, 0),
        reason="Bottle only added auth_basic in 0.9.0.")
requires_plugins = pytest.mark.skipif(version < (0, 9, 0),
        reason="Bottle only added auth_basic in 0.9.0.")

_test_application_index_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:index_page', 1)]

if version >= (0, 9, 0):
    _test_application_index_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_index_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_index_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

@validate_code_level_metrics("_target_application", "index_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_target_application:index_page',
        scoped_metrics=_test_application_index_scoped_metrics,
        custom_metrics=_test_application_index_custom_metrics)
def test_application_index(target_application):
    response = target_application.get('/index')
    response.mustcontain('INDEX RESPONSE')

_test_application_error_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:error_page', 1)]

if version >= (0, 9, 0):
    _test_application_error_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_error_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_error_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

if six.PY3:
    _test_application_error_errors = ['builtins:RuntimeError']
else:
    _test_application_error_errors = ['exceptions:RuntimeError']

@validate_code_level_metrics("_target_application", "error_page")
@validate_transaction_errors(errors=_test_application_error_errors)
@validate_transaction_metrics('_target_application:error_page',
        scoped_metrics=_test_application_error_scoped_metrics,
        custom_metrics=_test_application_error_custom_metrics)
def test_application_error(target_application):
    response = target_application.get('/error', status=500, expect_errors=True)

_test_application_not_found_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:error404_page', 1)]

if version >= (0, 9, 0):
    _test_application_not_found_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_not_found_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_not_found_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

@validate_code_level_metrics("_target_application", "error404_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_target_application:error404_page',
        scoped_metrics=_test_application_not_found_scoped_metrics,
        custom_metrics=_test_application_not_found_custom_metrics)
def test_application_not_found(target_application):
    response = target_application.get('/missing', status=404)
    response.mustcontain('NOT FOUND')

_test_application_auth_basic_fail_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:auth_basic_page', 1)]

if version >= (0, 9, 0):
    _test_application_auth_basic_fail_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_auth_basic_fail_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_auth_basic_fail_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

@requires_auth_basic
@validate_code_level_metrics("_target_application", "auth_basic_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_target_application:auth_basic_page',
        scoped_metrics=_test_application_auth_basic_fail_scoped_metrics,
        custom_metrics=_test_application_auth_basic_fail_custom_metrics)
def test_application_auth_basic_fail(target_application):
    response = target_application.get('/auth', status=401)

_test_application_auth_basic_okay_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:auth_basic_page', 1)]

if version >= (0, 9, 0):
    _test_application_auth_basic_okay_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_auth_basic_okay_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_auth_basic_okay_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

@requires_auth_basic
@validate_code_level_metrics("_target_application", "auth_basic_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_target_application:auth_basic_page',
        scoped_metrics=_test_application_auth_basic_okay_scoped_metrics,
        custom_metrics=_test_application_auth_basic_okay_custom_metrics)
def test_application_auth_basic_okay(target_application):
    authorization_value = base64.b64encode(b'user:password')
    if six.PY3:
        authorization_value = authorization_value.decode('Latin-1')
    environ = { 'HTTP_AUTHORIZATION': 'Basic ' + authorization_value }
    response = target_application.get('/auth', extra_environ=environ)
    response.mustcontain('AUTH OKAY')

_test_application_plugin_error_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_target_application:plugin_error_page', 1)]

if version >= (0, 9, 0):
    _test_application_plugin_error_scoped_metrics.extend([
        ('Function/bottle:Bottle.wsgi', 1)])
else:
    _test_application_plugin_error_scoped_metrics.extend([
        ('Function/bottle:Bottle.__call__', 1)])

_test_application_plugin_error_custom_metrics = [
        ('Python/Framework/Bottle/%s.%s.%s' % version, 1)]

@requires_plugins
@validate_code_level_metrics("_target_application", "plugin_error_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_target_application:plugin_error_page',
        scoped_metrics=_test_application_plugin_error_scoped_metrics,
        custom_metrics=_test_application_plugin_error_custom_metrics)
@override_ignore_status_codes([403])
def test_application_plugin_error_ignore(target_application):
    response = target_application.get('/plugin_error', status=403,
            expect_errors=True)

@requires_plugins
@validate_code_level_metrics("_target_application", "plugin_error_page")
@validate_transaction_errors(errors=['bottle:HTTPError'])
@validate_transaction_metrics('_target_application:plugin_error_page',
        scoped_metrics=_test_application_plugin_error_scoped_metrics,
        custom_metrics=_test_application_plugin_error_custom_metrics)
def test_application_plugin_error_capture(target_application):
    import newrelic.agent
    response = target_application.get('/plugin_error', status=403,
            expect_errors=True)

_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}

@override_application_settings(_test_html_insertion_settings)
def test_html_insertion(target_application):
    response = target_application.get('/html_insertion')

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')




# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from newrelic.core.agent import Agent
from newrelic.core.config import finalize_application_settings
from testing_support.fixtures import override_generic_settings


class FakeApplication(object):
    name = 'Fake'

    def __init__(self, *args, **kwargs):
        self.harvest_flexible = 0
        self.harvest_default = 0
        self.is_alive = True

    def harvest(self, shutdown=False, flexible=False, *args, **kwargs):
        assert self.is_alive

        if flexible:
            self.harvest_flexible += 1
        else:
            self.harvest_default += 1

        if shutdown:
            self.is_alive = False


class FakeAgent(Agent):
    def __init__(self, *args, **kwargs):
        super(FakeAgent, self).__init__(*args, **kwargs)
        self._applications = {'fake': FakeApplication()}


SETTINGS = finalize_application_settings({
    'enabled': True,
    'debug.disable_harvest_until_shutdown': False,
})


@pytest.fixture
def agent():
    agent = FakeAgent(SETTINGS)
    yield agent
    agent.shutdown_agent(timeout=5)
    assert not agent._harvest_thread.is_alive()


_override_settings = {
    'event_harvest_config.report_period_ms': 80.0 * 1000.0,
}


@override_generic_settings(SETTINGS, _override_settings)
def test_agent_final_harvest(agent):
    agent.activate_agent()
    assert agent._harvest_thread.is_alive()

    agent.shutdown_agent(timeout=5)
    assert not agent._harvest_thread.is_alive()

    assert agent._applications['fake'].harvest_flexible == 1
    assert agent._applications['fake'].harvest_default == 1



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import transient_function_wrapper

from testing_support.fixtures import catch_background_exceptions


def validate_database_trace_inputs(sql_parameters_type):

    @transient_function_wrapper('newrelic.api.database_trace',
            'DatabaseTrace.__init__')
    @catch_background_exceptions
    def _validate_database_trace_inputs(wrapped, instance, args, kwargs):
        def _bind_params(sql, dbapi2_module=None,
                connect_params=None, cursor_params=None, sql_parameters=None,
                execute_params=None, host=None, port_path_or_id=None,
                database_name=None, source=None):
            return (sql, dbapi2_module, connect_params,
                    cursor_params, sql_parameters, execute_params, source)

        (sql, dbapi2_module, connect_params, cursor_params,
            sql_parameters, execute_params, source) = _bind_params(*args, **kwargs)

        assert hasattr(dbapi2_module, 'connect')

        assert connect_params is None or isinstance(connect_params, tuple)

        if connect_params is not None:
            assert len(connect_params) == 2
            assert isinstance(connect_params[0], tuple)
            assert isinstance(connect_params[1], dict)

        assert cursor_params is None or isinstance(cursor_params, tuple)

        if cursor_params is not None:
            assert len(cursor_params) == 2
            assert isinstance(cursor_params[0], tuple)
            assert isinstance(cursor_params[1], dict)

        assert sql_parameters is None or isinstance(
                sql_parameters, sql_parameters_type)

        if execute_params is not None:
            assert len(execute_params) == 2
            assert isinstance(execute_params[0], tuple)
            assert isinstance(execute_params[1], dict)

        assert callable(source)

        return wrapped(*args, **kwargs)

    return _validate_database_trace_inputs



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from newrelic.api.transaction import ignore_transaction, end_of_transaction

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

from tasks import add, tsum


@validate_transaction_metrics(
        name='test_celery:test_celery_task_as_function_trace',
        scoped_metrics=[('Function/tasks.add', 1)],
        background_task=True)
@validate_code_level_metrics("tasks", "add")
@background_task()
def test_celery_task_as_function_trace():
    """
    Calling add() inside a transaction means the agent will record
    add() as a FunctionTrace.

    """
    result = add(3, 4)
    assert result == 7


@validate_transaction_metrics(
        name='tasks.add',
        group='Celery',
        scoped_metrics=[],
        background_task=True)
@validate_code_level_metrics("tasks", "add")
def test_celery_task_as_background_task():
    """
    Calling add() outside of a transaction means the agent will create
    a background transaction (with a group of 'Celery') and record add()
    as a background task.

    """
    result = add(3, 4)
    assert result == 7

@validate_transaction_metrics(
        name='test_celery:test_celery_tasks_multiple_function_traces',
        scoped_metrics=[('Function/tasks.add', 1),
                        ('Function/tasks.tsum', 1)],
        background_task=True)
@validate_code_level_metrics("tasks", "tsum")
@background_task()
def test_celery_tasks_multiple_function_traces():
    add_result = add(5, 6)
    assert add_result == 11

    tsum_result = tsum([1, 2, 3, 4])
    assert tsum_result == 10


@background_task()
def test_celery_tasks_ignore_transaction():
    """
    No transaction is recorded, due to the call to ignore_transaction(),
    so no validation fixture is used. The purpose of this test is to make
    sure the agent doesn't throw an error.

    """
    add_result = add(1, 2)
    assert add_result == 3

    ignore_transaction()

    tsum_result = tsum([1, 2, 3])
    assert tsum_result == 6


@validate_transaction_metrics(
        name='test_celery:test_celery_tasks_end_transaction',
        scoped_metrics=[('Function/tasks.add', 1)],
        background_task=True)
@background_task()
def test_celery_tasks_end_transaction():
    """
    Only functions that run before the call to end_of_transaction() are
    included in the transaction.

    """
    add_result = add(1, 2)
    assert add_result == 3

    end_of_transaction()

    tsum_result = tsum([1, 2, 3])
    assert tsum_result == 6



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from testing_support.db_settings import postgresql_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_database_trace_inputs import (
    validate_database_trace_inputs,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = postgresql_settings()[0]


@validate_transaction_metrics(
    "test_pyodbc:test_execute_via_cursor",
    scoped_metrics=[
        ("Function/pyodbc:connect", 1),
    ],
    rollup_metrics=[
        ("Datastore/all", 1),
        ("Datastore/allOther", 1),
        ("Datastore/ODBC/all", 1),
        ("Datastore/ODBC/allOther", 1),
    ],
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor(pyodbc_driver):
    import pyodbc

    with pyodbc.connect(
        "DRIVER={%s};SERVER=%s;PORT=%s;DATABASE=%s;UID=%s;PWD=%s"
        % (
            pyodbc_driver,
            DB_SETTINGS["host"],
            DB_SETTINGS["port"],
            DB_SETTINGS["name"],
            DB_SETTINGS["user"],
            DB_SETTINGS["password"],
        )
    ) as connection:
        cursor = connection.cursor()
        cursor.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])
        cursor.execute("""create table %s """ % DB_SETTINGS["table_name"] + """(a integer, b real, c text)""")
        cursor.executemany(
            """insert into %s """ % DB_SETTINGS["table_name"] + """values (?, ?, ?)""",
            [(1, 1.0, "1.0"), (2, 2.2, "2.2"), (3, 3.3, "3.3")],
        )
        cursor.execute("""select * from %s""" % DB_SETTINGS["table_name"])
        for row in cursor:
            pass
        cursor.execute(
            """update %s """ % DB_SETTINGS["table_name"] + """set a=?, b=?, c=? where a=?""",
            (4, 4.0, "4.0", 1),
        )
        cursor.execute("""delete from %s where a=2""" % DB_SETTINGS["table_name"])
        connection.commit()

        cursor.execute("SELECT now()")
        cursor.execute("SELECT pg_sleep(0.25)")

        connection.rollback()
        connection.commit()


@validate_transaction_metrics(
    "test_pyodbc:test_rollback_on_exception",
    scoped_metrics=[
        ("Function/pyodbc:connect", 1),
    ],
    rollup_metrics=[
        ("Datastore/all", 1),
        ("Datastore/allOther", 1),
        ("Datastore/ODBC/all", 1),
        ("Datastore/ODBC/allOther", 1),
    ],
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception(pyodbc_driver):
    import pyodbc

    with pytest.raises(RuntimeError):
        with pyodbc.connect(
            "DRIVER={%s};SERVER=%s;PORT=%s;DATABASE=%s;UID=%s;PWD=%s"
            % (
                pyodbc_driver,
                DB_SETTINGS["host"],
                DB_SETTINGS["port"],
                DB_SETTINGS["name"],
                DB_SETTINGS["user"],
                DB_SETTINGS["password"],
            )
        ) as connection:
            raise RuntimeError("error")


@pytest.fixture
def pyodbc_driver():
    import pyodbc

    driver_name = "PostgreSQL Unicode"
    assert driver_name in pyodbc.drivers()
    return driver_name



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
import sys

import pytest
from _test_code_level_metrics import (
    CLASS_INSTANCE,
    CLASS_INSTANCE_CALLABLE,
    TYPE_CONSTRUCTOR_CALLABLE_CLASS_INSTANCE,
    TYPE_CONSTRUCTOR_CLASS_INSTANCE,
    ExerciseClass,
    ExerciseClassCallable,
    ExerciseTypeConstructor,
    ExerciseTypeConstructorCallable,
)
from _test_code_level_metrics import __file__ as FILE_PATH
from _test_code_level_metrics import (
    exercise_function,
    exercise_lambda,
    exercise_partial,
)
from testing_support.fixtures import dt_enabled, override_application_settings
from testing_support.validators.validate_span_events import validate_span_events

import newrelic.packages.six as six
from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace

is_pypy = hasattr(sys, "pypy_version_info")

NAMESPACE = "_test_code_level_metrics"
CLASS_NAMESPACE = ".".join((NAMESPACE, "ExerciseClass"))
CALLABLE_CLASS_NAMESPACE = ".".join((NAMESPACE, "ExerciseClassCallable"))
TYPE_CONSTRUCTOR_NAMESPACE = ".".join((NAMESPACE, "ExerciseTypeConstructor"))
TYPE_CONSTRUCTOR_CALLABLE_NAMESPACE = ".".join((NAMESPACE, "ExerciseTypeConstructorCallable"))
FUZZY_NAMESPACE = CLASS_NAMESPACE if six.PY3 else NAMESPACE
if FILE_PATH.endswith(".pyc"):
    FILE_PATH = FILE_PATH[:-1]

SQLITE_CONNECTION = sqlite3.Connection(":memory:")

BUILTIN_ATTRS = {"code.filepath": "<builtin>", "code.lineno": None} if not is_pypy else {}


def merge_dicts(A, B):
    d = {}
    d.update(A)
    d.update(B)
    return d


@pytest.fixture
def extract():
    def _extract(obj):
        with FunctionTrace("_test", source=obj):
            pass

    return _extract


_TEST_BASIC_CALLABLES = {
    "function": (
        exercise_function,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_function",
            "code.lineno": 17,
            "code.namespace": NAMESPACE,
        },
    ),
    "lambda": (
        exercise_lambda,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "<lambda>",
            "code.lineno": 75,
            "code.namespace": NAMESPACE,
        },
    ),
    "partial": (
        exercise_partial,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_function",
            "code.lineno": 17,
            "code.namespace": NAMESPACE,
        },
    ),
    "builtin_function": (
        max,
        (1, 2),
        merge_dicts(
            {
                "code.function": "max",
                "code.namespace": "builtins" if six.PY3 else "__builtin__",
            },
            BUILTIN_ATTRS,
        ),
    ),
    "builtin_module_function": (
        sqlite3.connect,
        (":memory:",),
        merge_dicts(
            {
                "code.function": "connect",
                "code.namespace": "_sqlite3",
            },
            BUILTIN_ATTRS,
        ),
    ),
}


@pytest.mark.parametrize(
    "func,args,agents",
    [pytest.param(*args, id=id_) for id_, args in six.iteritems(_TEST_BASIC_CALLABLES)],
)
def test_code_level_metrics_basic_callables(func, args, agents, extract):
    @override_application_settings(
        {
            "code_level_metrics.enabled": True,
        }
    )
    @dt_enabled
    @validate_span_events(
        count=1,
        exact_agents=agents,
    )
    @background_task()
    def _test():
        extract(func)

    _test()


_TEST_METHODS = {
    "method": (
        CLASS_INSTANCE.exercise_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_method",
            "code.lineno": 22,
            "code.namespace": CLASS_NAMESPACE,
        },
    ),
    "static_method": (
        CLASS_INSTANCE.exercise_static_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_static_method",
            "code.lineno": 25,
            "code.namespace": FUZZY_NAMESPACE,
        },
    ),
    "class_method": (
        ExerciseClass.exercise_class_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_class_method",
            "code.lineno": 29,
            "code.namespace": CLASS_NAMESPACE,
        },
    ),
    "call_method": (
        CLASS_INSTANCE_CALLABLE,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "__call__",
            "code.lineno": 35,
            "code.namespace": CALLABLE_CLASS_NAMESPACE,
        },
    ),
    "builtin_method": (
        SQLITE_CONNECTION.__enter__,
        (),
        merge_dicts(
            {
                "code.function": "__enter__",
                "code.namespace": "sqlite3.Connection" if not is_pypy else "_sqlite3.Connection",
            },
            BUILTIN_ATTRS,
        ),
    ),
}


@pytest.mark.parametrize(
    "func,args,agents",
    [pytest.param(*args, id=id_) for id_, args in six.iteritems(_TEST_METHODS)],
)
def test_code_level_metrics_methods(func, args, agents, extract):
    @override_application_settings(
        {
            "code_level_metrics.enabled": True,
        }
    )
    @dt_enabled
    @validate_span_events(
        count=1,
        exact_agents=agents,
    )
    @background_task()
    def _test():
        extract(func)

    _test()


_TEST_TYPE_CONSTRUCTOR_METHODS = {
    "method": (
        TYPE_CONSTRUCTOR_CLASS_INSTANCE.exercise_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_method",
            "code.lineno": 39,
            "code.namespace": TYPE_CONSTRUCTOR_NAMESPACE,
        },
    ),
    "static_method": (
        TYPE_CONSTRUCTOR_CLASS_INSTANCE.exercise_static_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_static_method",
            "code.lineno": 43,
            "code.namespace": NAMESPACE,
        },
    ),
    "class_method": (
        ExerciseTypeConstructor.exercise_class_method,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "exercise_class_method",
            "code.lineno": 48,
            "code.namespace": TYPE_CONSTRUCTOR_NAMESPACE,
        },
    ),
    "lambda_method": (
        ExerciseTypeConstructor.exercise_lambda,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "<lambda>",
            "code.lineno": 61,
            # Lambdas behave strangely in type constructors on Python 2 and use the class namespace.
            "code.namespace": NAMESPACE if six.PY3 else TYPE_CONSTRUCTOR_NAMESPACE,
        },
    ),
    "call_method": (
        TYPE_CONSTRUCTOR_CALLABLE_CLASS_INSTANCE,
        (),
        {
            "code.filepath": FILE_PATH,
            "code.function": "__call__",
            "code.lineno": 53,
            "code.namespace": TYPE_CONSTRUCTOR_CALLABLE_NAMESPACE,
        },
    ),
}


@pytest.mark.parametrize(
    "func,args,agents",
    [pytest.param(*args, id=id_) for id_, args in six.iteritems(_TEST_TYPE_CONSTRUCTOR_METHODS)],
)
def test_code_level_metrics_type_constructor_methods(func, args, agents, extract):
    @override_application_settings(
        {
            "code_level_metrics.enabled": True,
        }
    )
    @dt_enabled
    @validate_span_events(
        count=1,
        exact_agents=agents,
    )
    @background_task()
    def _test():
        extract(func)

    _test()


_TEST_OBJECTS = {
    "class": (
        ExerciseClass,
        {
            "code.filepath": FILE_PATH,
            "code.function": "ExerciseClass",
            "code.lineno": 21,
            "code.namespace": NAMESPACE,
        },
    ),
    "callable_class": (
        ExerciseClassCallable,
        {
            "code.filepath": FILE_PATH,
            "code.function": "ExerciseClassCallable",
            "code.lineno": 34,
            "code.namespace": NAMESPACE,
        },
    ),
    "type_constructor_class": (
        ExerciseTypeConstructor,
        {
            "code.filepath": FILE_PATH,
            "code.function": "ExerciseTypeConstructor",
            "code.namespace": NAMESPACE,
        },
    ),
    "type_constructor_class_callable_class": (
        ExerciseTypeConstructorCallable,
        {
            "code.filepath": FILE_PATH,
            "code.function": "ExerciseTypeConstructorCallable",
            "code.namespace": NAMESPACE,
        },
    ),
    "non_callable_object": (
        CLASS_INSTANCE,
        {
            "code.filepath": FILE_PATH,
            "code.function": "ExerciseClass",
            "code.lineno": 21,
            "code.namespace": NAMESPACE,
        },
    ),
}


@pytest.mark.parametrize(
    "obj,agents",
    [pytest.param(*args, id=id_) for id_, args in six.iteritems(_TEST_OBJECTS)],
)
def test_code_level_metrics_objects(obj, agents, extract):
    @override_application_settings(
        {
            "code_level_metrics.enabled": True,
        }
    )
    @dt_enabled
    @validate_span_events(
        count=1,
        exact_agents=agents,
    )
    @background_task()
    def _test():
        extract(obj)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json

import pytest
import webtest
from testing_support.fixtures import override_application_settings, validate_attributes
from testing_support.validators.validate_error_event_attributes import (
    validate_error_event_attributes,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance
from newrelic.api.background_task import BackgroundTask, background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import (
    accept_distributed_trace_headers,
    accept_distributed_trace_payload,
    create_distributed_trace_payload,
    current_span_id,
    current_trace_id,
    current_transaction,
)
from newrelic.api.web_transaction import WSGIWebTransaction
from newrelic.api.wsgi_application import wsgi_application

distributed_trace_intrinsics = ["guid", "traceId", "priority", "sampled"]
inbound_payload_intrinsics = [
    "parent.type",
    "parent.app",
    "parent.account",
    "parent.transportType",
    "parent.transportDuration",
]

payload = {
    "v": [0, 1],
    "d": {
        "ac": "1",
        "ap": "2827902",
        "id": "7d3efb1b173fecfa",
        "pa": "5e5733a911cfbc73",
        "pr": 10.001,
        "sa": True,
        "ti": 1518469636035,
        "tr": "d6b4ba0c3a712ca",
        "ty": "App",
    },
}
parent_order = ["parent_type", "parent_account", "parent_app", "parent_transport_type"]
parent_info = {
    "parent_type": payload["d"]["ty"],
    "parent_account": payload["d"]["ac"],
    "parent_app": payload["d"]["ap"],
    "parent_transport_type": "HTTP",
}


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"
    output = b"hello world"
    response_headers = [("Content-type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]

    txn = current_transaction()

    # Make assertions on the WSGIWebTransaction object
    assert txn._distributed_trace_state
    assert txn.parent_type == "App"
    assert txn.parent_app == "2827902"
    assert txn.parent_account == "1"
    assert txn.parent_span == "7d3efb1b173fecfa"
    assert txn.parent_transport_type == "HTTP"
    assert isinstance(txn.parent_transport_duration, float)
    assert txn._trace_id == "d6b4ba0c3a712ca"
    assert txn.priority == 10.001
    assert txn.sampled

    start_response(status, response_headers)
    return [output]


test_application = webtest.TestApp(target_wsgi_application)

_override_settings = {
    "trusted_account_key": "1",
    "distributed_tracing.enabled": True,
}


_metrics = [
    ("Supportability/DistributedTrace/AcceptPayload/Success", 1),
    ("Supportability/TraceContext/Accept/Success", None),
]


@override_application_settings(_override_settings)
@validate_transaction_metrics("", group="Uri", rollup_metrics=_metrics)
def test_distributed_tracing_web_transaction():
    headers = {"newrelic": json.dumps(payload)}
    response = test_application.get("/", headers=headers)
    assert "X-NewRelic-App-Data" not in response.headers


@pytest.mark.parametrize("span_events", (True, False))
@pytest.mark.parametrize("accept_payload", (True, False))
def test_distributed_trace_attributes(span_events, accept_payload):
    if accept_payload:
        _required_intrinsics = distributed_trace_intrinsics + inbound_payload_intrinsics
        _forgone_txn_intrinsics = []
        _forgone_error_intrinsics = []
        _exact_intrinsics = {
            "parent.type": "Mobile",
            "parent.app": "2827902",
            "parent.account": "1",
            "parent.transportType": "HTTP",
            "traceId": "d6b4ba0c3a712ca",
        }
        _exact_txn_attributes = {"agent": {}, "user": {}, "intrinsic": _exact_intrinsics.copy()}
        _exact_error_attributes = {"agent": {}, "user": {}, "intrinsic": _exact_intrinsics.copy()}
        _exact_txn_attributes["intrinsic"]["parentId"] = "7d3efb1b173fecfa"
        _exact_txn_attributes["intrinsic"]["parentSpanId"] = "c86df80de2e6f51c"

        _forgone_error_intrinsics.append("parentId")
        _forgone_error_intrinsics.append("parentSpanId")
        _forgone_txn_intrinsics.append("grandparentId")
        _forgone_error_intrinsics.append("grandparentId")

        _required_attributes = {"intrinsic": _required_intrinsics, "agent": [], "user": []}
        _forgone_txn_attributes = {"intrinsic": _forgone_txn_intrinsics, "agent": [], "user": []}
        _forgone_error_attributes = {"intrinsic": _forgone_error_intrinsics, "agent": [], "user": []}
    else:
        _required_intrinsics = distributed_trace_intrinsics
        _forgone_txn_intrinsics = _forgone_error_intrinsics = inbound_payload_intrinsics + [
            "grandparentId",
            "parentId",
            "parentSpanId",
        ]

        _required_attributes = {"intrinsic": _required_intrinsics, "agent": [], "user": []}
        _forgone_txn_attributes = {"intrinsic": _forgone_txn_intrinsics, "agent": [], "user": []}
        _forgone_error_attributes = {"intrinsic": _forgone_error_intrinsics, "agent": [], "user": []}
        _exact_txn_attributes = _exact_error_attributes = None

    _forgone_trace_intrinsics = _forgone_error_intrinsics

    test_settings = _override_settings.copy()
    test_settings["span_events.enabled"] = span_events

    @override_application_settings(test_settings)
    @validate_transaction_event_attributes(_required_attributes, _forgone_txn_attributes, _exact_txn_attributes)
    @validate_error_event_attributes(_required_attributes, _forgone_error_attributes, _exact_error_attributes)
    @validate_attributes("intrinsic", _required_intrinsics, _forgone_trace_intrinsics)
    @background_task(name="test_distributed_trace_attributes")
    def _test():
        txn = current_transaction()

        payload = {
            "v": [0, 1],
            "d": {
                "ty": "Mobile",
                "ac": "1",
                "ap": "2827902",
                "id": "c86df80de2e6f51c",
                "tr": "d6b4ba0c3a712ca",
                "ti": 1518469636035,
                "tx": "7d3efb1b173fecfa",
            },
        }
        payload["d"]["pa"] = "5e5733a911cfbc73"

        if accept_payload:
            result = accept_distributed_trace_payload(payload)
            assert result
        else:
            create_distributed_trace_payload()

        try:
            raise ValueError("cookies")
        except ValueError:
            txn.notice_error()

    _test()


_forgone_attributes = {
    "agent": [],
    "user": [],
    "intrinsic": (inbound_payload_intrinsics + ["grandparentId"]),
}


@override_application_settings(_override_settings)
@validate_transaction_event_attributes({}, _forgone_attributes)
@validate_error_event_attributes({}, _forgone_attributes)
@validate_attributes("intrinsic", {}, _forgone_attributes["intrinsic"])
@background_task(name="test_distributed_trace_attrs_omitted")
def test_distributed_trace_attrs_omitted():
    txn = current_transaction()
    try:
        raise ValueError("cookies")
    except ValueError:
        txn.notice_error()


# test our distributed_trace metrics by creating a transaction and then forcing
# it to process a distributed trace payload
@pytest.mark.parametrize("web_transaction", (True, False))
@pytest.mark.parametrize("gen_error", (True, False))
@pytest.mark.parametrize("has_parent", (True, False))
def test_distributed_tracing_metrics(web_transaction, gen_error, has_parent):
    def _make_dt_tag(pi):
        return "%s/%s/%s/%s/all" % tuple(pi[x] for x in parent_order)

    # figure out which metrics we'll see based on the test params
    # note: we'll always see DurationByCaller if the distributed
    # tracing flag is turned on
    metrics = ["DurationByCaller"]
    if gen_error:
        metrics.append("ErrorsByCaller")
    if has_parent:
        metrics.append("TransportDuration")

    tag = None
    dt_payload = copy.deepcopy(payload)

    # if has_parent is True, our metric name will be info about the parent,
    # otherwise it is Unknown/Unknown/Unknown/Unknown
    if has_parent:
        tag = _make_dt_tag(parent_info)
    else:
        # tag = _make_dt_tag(dict((x, "Unknown") for x in parent_order))
        tag = _make_dt_tag(dict((x, "Unknown") for x in parent_info.keys()))
        del dt_payload["d"]["tr"]

    # now run the test
    transaction_name = "test_dt_metrics_%s" % "_".join(metrics)
    _rollup_metrics = [
        ("%s/%s%s" % (x, tag, bt), 1) for x in metrics for bt in ["", "Web" if web_transaction else "Other"]
    ]

    def _make_test_transaction():
        application = application_instance()

        if not web_transaction:
            return BackgroundTask(application, transaction_name)

        environ = {"REQUEST_URI": "/trace_ends_after_txn"}
        tn = WSGIWebTransaction(application, environ)
        tn.set_transaction_name(transaction_name)
        return tn

    @override_application_settings(_override_settings)
    @validate_transaction_metrics(
        transaction_name, background_task=not (web_transaction), rollup_metrics=_rollup_metrics
    )
    def _test():
        with _make_test_transaction() as transaction:
            transaction.accept_distributed_trace_payload(dt_payload)

            if gen_error:
                try:
                    1 / 0
                except ZeroDivisionError:
                    transaction.notice_error()

    _test()


NEW_RELIC_ACCEPTED = [
    ("Supportability/DistributedTrace/AcceptPayload/Success", 1),
    ("Supportability/TraceContext/Accept/Success", None),
    ("Supportability/TraceContext/TraceParent/Accept/Success", None),
    ("Supportability/TraceContext/Accept/Success", None),
]
TRACE_CONTEXT_ACCEPTED = [
    ("Supportability/TraceContext/Accept/Success", 1),
    ("Supportability/TraceContext/TraceParent/Accept/Success", 1),
    ("Supportability/TraceContext/Accept/Success", 1),
    ("Supportability/DistributedTrace/AcceptPayload/Success", None),
]
NO_HEADERS_ACCEPTED = [
    ("Supportability/DistributedTrace/AcceptPayload/Success", None),
    ("Supportability/TraceContext/Accept/Success", None),
    ("Supportability/TraceContext/TraceParent/Accept/Success", None),
    ("Supportability/TraceContext/Accept/Success", None),
]
TRACEPARENT = "00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01"
TRACESTATE = "rojo=f06a0ba902b7,congo=t61rcWkgMzE"


@pytest.mark.parametrize(
    "traceparent,tracestate,newrelic,metrics",
    [
        (False, False, False, NO_HEADERS_ACCEPTED),
        (False, False, True, NEW_RELIC_ACCEPTED),
        (False, True, True, NEW_RELIC_ACCEPTED),
        (False, True, False, NO_HEADERS_ACCEPTED),
        (True, True, True, TRACE_CONTEXT_ACCEPTED),
        (True, False, False, TRACE_CONTEXT_ACCEPTED),
        (True, False, True, TRACE_CONTEXT_ACCEPTED),
        (True, True, False, TRACE_CONTEXT_ACCEPTED),
    ],
)
@override_application_settings(_override_settings)
def test_distributed_tracing_backwards_compatibility(traceparent, tracestate, newrelic, metrics):
    headers = []
    if traceparent:
        headers.append(("traceparent", TRACEPARENT))
    if tracestate:
        headers.append(("tracestate", TRACESTATE))
    if newrelic:
        headers.append(("newrelic", json.dumps(payload)))

    @validate_transaction_metrics(
        "test_distributed_tracing_backwards_compatibility", background_task=True, rollup_metrics=metrics
    )
    @background_task(name="test_distributed_tracing_backwards_compatibility")
    def _test():
        accept_distributed_trace_headers(headers)

    _test()


@background_task(name="test_current_trace_id_api_inside_transaction")
def test_current_trace_id_api_inside_transaction():
    trace_id = current_trace_id()
    assert len(trace_id) == 32
    assert trace_id == current_transaction().trace_id


def test_current_trace_id_api_outside_transaction():
    trace_id = current_trace_id()
    assert trace_id is None


@background_task(name="test_current_span_id_api_inside_transaction")
def test_current_span_id_inside_transaction():
    span_id = current_span_id()
    assert span_id == current_trace().guid


def test_current_span_id_outside_transaction():
    span_id = current_span_id()
    assert span_id is None


@pytest.mark.parametrize("trusted_account_key", ("1", None), ids=("tk_set", "tk_unset"))
def test_outbound_dt_payload_generation(trusted_account_key):
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "account_id": "1",
            "trusted_account_key": trusted_account_key,
            "primary_application_id": "1",
        }
    )
    @background_task(name="test_outbound_dt_payload_generation")
    def _test_outbound_dt_payload_generation():
        transaction = current_transaction()
        payload = ExternalTrace.generate_request_headers(transaction)
        if trusted_account_key:
            assert payload
            # Ensure trusted account key present as vendor
            assert dict(payload)["tracestate"].startswith("1@nr=")
        else:
            assert not payload

    _test_outbound_dt_payload_generation()


@pytest.mark.parametrize("trusted_account_key", ("1", None), ids=("tk_set", "tk_unset"))
def test_inbound_dt_payload_acceptance(trusted_account_key):
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "account_id": "1",
            "trusted_account_key": trusted_account_key,
            "primary_application_id": "1",
        }
    )
    @background_task(name="_test_inbound_dt_payload_acceptance")
    def _test_inbound_dt_payload_acceptance():
        transaction = current_transaction()

        payload = {
            "v": [0, 1],
            "d": {
                "ty": "Mobile",
                "ac": "1",
                "tk": "1",
                "ap": "2827902",
                "pa": "5e5733a911cfbc73",
                "id": "7d3efb1b173fecfa",
                "tr": "d6b4ba0c3a712ca",
                "ti": 1518469636035,
                "tx": "8703ff3d88eefe9d",
            },
        }

        result = transaction.accept_distributed_trace_payload(payload)
        if trusted_account_key:
            assert result
        else:
            assert not result

    _test_inbound_dt_payload_acceptance()
